import json

from utils.expected_Hstrain_functions import (
    expected_Hstrain_functions_fcc_1,
    expected_Hstrain_functions_hcp_1,
    expected_Hstrain_functions_lowsym_1,
)
from utils.helpers import (
    assert_expected_cluster_functions_detailed,
    assert_expected_functions,
)

import libcasm.configuration as casmconfig
import libcasm.xtal as xtal
import libcasm.xtal.prims as xtal_prims
from casm.bset import (
    build_cluster_functions,
)
from casm.bset.cluster_functions import (
    make_global_dof_matrix_rep,
)
from casm.bset.polynomial_functions import (
    Variable,
    make_symmetry_adapted_polynomials,
)


def test_Hstrain_fcc_1(session_shared_datadir):
    key = "Hstrain"
    xtal_prim = xtal_prims.FCC(
        r=0.5,
        occ_dof=["A"],
        global_dof=[xtal.DoFSetBasis(key)],
    )
    # print(xtal.pretty_json(xtal_prim.to_dict()))

    # test directly using make_symmetry_adapted_polynomials
    prim = casmconfig.Prim(xtal_prim)
    matrix_rep = make_global_dof_matrix_rep(
        prim=prim,
        key=key,
    )
    basis_set = make_symmetry_adapted_polynomials(
        matrix_rep=matrix_rep,
        variables=[
            Variable(name="{E_{xx}}", key=key),
            Variable(name="{E_{yy}}", key=key),
            Variable(name="{E_{zz}}", key=key),
            Variable(name="{\\sqrt{2}E_{yz}}", key=key),
            Variable(name="{\\sqrt{2}E_{xz}}", key=key),
            Variable(name="{\\sqrt{2}E_{xy}}", key=key),
        ],
        variable_subsets=[[0, 1, 2, 3, 4, 5]],
        min_poly_order=1,
        max_poly_order=3,
    )
    assert len(basis_set) == 10
    # print_expected_functions(basis_set)
    expected = expected_Hstrain_functions_fcc_1()
    assert_expected_functions(basis_set, expected)

    # test build_cluster_functions with only global DoF
    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {},
            "basis_function_specs": {
                "global_max_poly_order": 3,
            },
        },
    )
    functions, clusters = (builder.functions, builder.clusters)

    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_Hstrain_functions_fcc_1.json",
    # )
    with open(session_shared_datadir / "expected_Hstrain_functions_fcc_1.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))

    assert_expected_functions(functions[0][0], expected)


def test_Hstrain_fcc_2(session_shared_datadir):
    """Test generating only strain polynomials when the prim also has occ DoF"""
    key = "Hstrain"
    xtal_prim = xtal_prims.FCC(
        r=0.5,
        occ_dof=["A", "B"],
        global_dof=[xtal.DoFSetBasis(key)],
    )
    # print(xtal.pretty_json(xtal_prim.to_dict()))

    # test build_cluster_functions with only global DoF
    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dofs": ["Hstrain"],
                "global_max_poly_order": 3,
            },
        },
    )
    functions, clusters = (builder.functions, builder.clusters)

    assert len(clusters) == 4
    assert len(clusters[0]) == 1
    assert len(clusters[1]) == 1
    assert len(clusters[2]) == 6
    assert len(clusters[3]) == 8

    assert len(functions) == 4
    assert len(functions[0]) == 1  # null cluster
    assert len(functions[0][0]) == 10
    assert len(functions[1]) == 1  # point cluster
    assert len(functions[1][0]) == 0
    assert len(functions[2]) == 6  # pair cluster
    assert len(functions[2][0]) == 0
    assert len(functions[3]) == 8  # triplet cluster
    assert len(functions[3][0]) == 0

    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_Hstrain_functions_fcc_2.json",
    # )
    with open(session_shared_datadir / "expected_Hstrain_functions_fcc_2.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))

    expected = expected_Hstrain_functions_fcc_1()
    assert_expected_functions(functions[0][0], expected)


def test_Hstrain_fcc_3(session_shared_datadir):
    """Test generating only strain polynomials when the prim also has disp DoF"""
    key = "Hstrain"
    xtal_prim = xtal_prims.FCC(
        r=0.5,
        occ_dof=["A"],
        local_dof=[xtal.DoFSetBasis("disp")],
        global_dof=[xtal.DoFSetBasis(key)],
    )
    # print(xtal.pretty_json(xtal_prim.to_dict()))

    # test build_cluster_functions with only global DoF
    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dofs": ["Hstrain"],
                "global_max_poly_order": 3,
            },
        },
    )
    functions, clusters = (builder.functions, builder.clusters)

    assert len(clusters) == 4
    assert len(clusters[0]) == 1
    assert len(clusters[1]) == 1
    assert len(clusters[2]) == 6
    assert len(clusters[3]) == 8

    assert len(functions) == 4
    assert len(functions[0]) == 1  # null cluster
    assert len(functions[0][0]) == 10
    assert len(functions[1]) == 1  # point cluster
    assert len(functions[1][0]) == 0
    assert len(functions[2]) == 6  # pair cluster
    assert len(functions[2][0]) == 0
    assert len(functions[3]) == 8  # triplet cluster
    assert len(functions[3][0]) == 0

    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_Hstrain_functions_fcc_3.json",
    # )
    with open(session_shared_datadir / "expected_Hstrain_functions_fcc_3.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))

    expected = expected_Hstrain_functions_fcc_1()
    assert_expected_functions(functions[0][0], expected)


def test_Hstrain_hcp_1(session_shared_datadir):
    key = "Hstrain"
    xtal_prim = xtal_prims.HCP(
        r=0.5,
        occ_dof=["A"],
        global_dof=[xtal.DoFSetBasis(key)],
    )
    # print(xtal.pretty_json(xtal_prim.to_dict()))

    # test directly using make_symmetry_adapted_polynomials
    prim = casmconfig.Prim(xtal_prim)
    matrix_rep = make_global_dof_matrix_rep(
        prim=prim,
        key=key,
    )
    basis_set = make_symmetry_adapted_polynomials(
        matrix_rep=matrix_rep,
        variables=[
            Variable(name="{E_{xx}}", key=key),
            Variable(name="{E_{yy}}", key=key),
            Variable(name="{E_{zz}}", key=key),
            Variable(name="{\\sqrt{2}E_{yz}}", key=key),
            Variable(name="{\\sqrt{2}E_{xz}}", key=key),
            Variable(name="{\\sqrt{2}E_{xy}}", key=key),
        ],
        variable_subsets=[[0, 1, 2, 3, 4, 5]],
        min_poly_order=1,
        max_poly_order=3,
    )
    assert len(basis_set) == 17

    # print_expected_functions(basis_set)
    expected = expected_Hstrain_functions_hcp_1()
    assert_expected_functions(basis_set, expected)

    # test build_cluster_functions with only global DoF
    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {},
            "basis_function_specs": {
                "dofs": ["Hstrain"],
                "global_max_poly_order": 3,
            },
        },
    )
    functions, clusters = (builder.functions, builder.clusters)

    assert len(clusters) == 1
    assert len(clusters[0]) == 1  # null cluster

    assert len(functions) == 1
    assert len(functions[0]) == 1  # null cluster
    assert len(functions[0][0]) == 17

    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_Hstrain_functions_hcp_1.json",
    # )
    with open(session_shared_datadir / "expected_Hstrain_functions_hcp_1.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))

    assert_expected_functions(functions[0][0], expected)


def test_Hstrain_lowsym_1(lowsym_Hstrain_prim, session_shared_datadir):
    key = "Hstrain"
    xtal_prim = lowsym_Hstrain_prim
    # print(xtal.pretty_json(xtal_prim.to_dict()))

    # test directly using make_symmetry_adapted_polynomials
    prim = casmconfig.Prim(xtal_prim)
    factor_group_elements = prim.factor_group.elements
    assert len(factor_group_elements) == 1
    matrix_rep = make_global_dof_matrix_rep(
        prim=prim,
        key=key,
    )
    basis_set = make_symmetry_adapted_polynomials(
        matrix_rep=matrix_rep,
        variables=[
            Variable(name="{E_{xx}}", key=key),
            Variable(name="{E_{yy}}", key=key),
            Variable(name="{E_{zz}}", key=key),
            Variable(name="{\\sqrt{2}E_{yz}}", key=key),
            Variable(name="{\\sqrt{2}E_{xz}}", key=key),
            Variable(name="{\\sqrt{2}E_{xy}}", key=key),
        ],
        variable_subsets=[[0, 1, 2, 3, 4, 5]],
        min_poly_order=1,
        max_poly_order=3,
    )
    assert len(basis_set) == 83

    # print_expected_functions(basis_set)
    expected = expected_Hstrain_functions_lowsym_1()
    assert_expected_functions(basis_set, expected)

    # test build_cluster_functions with only global DoF
    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {},
            "basis_function_specs": {
                "dofs": ["Hstrain"],
                "global_max_poly_order": 3,
            },
        },
    )
    functions, clusters = (builder.functions, builder.clusters)

    assert len(clusters) == 1
    assert len(clusters[0]) == 1  # null cluster

    assert len(functions) == 1
    assert len(functions[0]) == 1  # null cluster
    assert len(functions[0][0]) == 83

    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_Hstrain_functions_lowsym_1.json",
    # )
    with open(session_shared_datadir / "expected_Hstrain_functions_lowsym_1.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))

    assert_expected_functions(functions[0][0], expected)



import pathlib
from typing import Any, Optional

import numpy as np

import libcasm.xtal as xtal


def print_factor_group(prim):
    factor_group_elements = prim.factor_group.elements
    i = 1
    for op in factor_group_elements:
        syminfo = xtal.SymInfo(op, prim.lattice())
        print(str(i) + ":", syminfo.brief_cart())
        i += 1


def print_expected_functions(functions):
    _expected = []
    for i, f in enumerate(functions):
        # print(f"~~~ order: {f.order()}, function_index: {i} ~~~")
        # f._basic_print()
        _data = f.to_dict()
        _expected.append(
            {
                "coeff_coords": _data["coeff_coords"],
                "coeff_data": _data["coeff_data"],
            }
        )
    print(xtal.pretty_json(_expected))


def assert_expected_functions(functions, expected):
    assert len(functions) == len(expected)
    n_found = 0
    for g in expected:
        found = False
        expected_coords = np.array(g["coeff_coords"]).transpose()
        expected_data = np.array(g["coeff_data"])
        for f in functions:
            if (
                f.coeff.coords.shape == expected_coords.shape
                and np.allclose(f.coeff.coords.tolist(), expected_coords)
                and f.coeff.data.shape == expected_data.shape
                and np.allclose(f.coeff.data.tolist(), expected_data)
            ):
                found = True
                break
        if not found:
            print("expected function not found:\n", xtal.pretty_json(g))
            print("found functions:")
            for i, x in enumerate(functions):
                print(f"function {i}:\n", xtal.pretty_json(x.to_dict()))
        assert found
        n_found += 1
    assert n_found == len(functions)


def print_expected_cluster_functions(functions):
    """Print functions on the prototype clusters

    Parameters
    ----------
    functions: list[list[list[PolynomialFunction]]]
        As generated by :class:`ClusterFunctionsBuilder`.

    """
    _expected_by_orbit = []
    for i_orbit, orbit_functions in enumerate(functions):
        _expected = []
        prototype_cluster_functions = orbit_functions[0]
        for i, f in enumerate(prototype_cluster_functions):
            # print(f"~~~ order: {f.order()}, function_index: {i} ~~~")
            # f._basic_print()
            _data = f.to_dict()
            _expected.append(
                {
                    "coeff_coords": _data["coeff_coords"],
                    "coeff_data": _data["coeff_data"],
                }
            )
        _expected_by_orbit.append(_expected)
    print(xtal.pretty_json(_expected_by_orbit))


def assert_expected_cluster_functions(
    functions,
    expected_by_orbit,
):
    assert len(functions) == len(expected_by_orbit)
    for i_orbit in range(len(functions)):
        prototype_cluster_functions = functions[i_orbit][0]
        assert_expected_functions(
            prototype_cluster_functions,
            expected_by_orbit[i_orbit],
        )


def print_expected_cluster_functions_detailed(
    functions, file: Optional[pathlib.Path] = None
):
    """Print functions on the prototype clusters

    Parameters
    ----------
    functions: list[list[list[PolynomialFunction]]]
        As generated by :class:`ClusterFunctionsBuilder`.

    """

    data = {"functions": []}

    for i_orbit, orbit in enumerate(functions):
        orbit_data = []
        for i_equiv, equiv in enumerate(orbit):
            equiv_data = []
            for i_func, func in enumerate(equiv):
                d = func.to_dict()
                d["i_orbit"] = i_orbit
                d["i_equiv"] = i_equiv
                d["i_func"] = i_func
                equiv_data.append(d)
            orbit_data.append(equiv_data)
        data["functions"].append(orbit_data)

    if file is not None:
        with open(file, "w") as f:
            f.write(xtal.pretty_json(data))
    else:
        print(xtal.pretty_json(data))


def assert_expected_cluster_functions_detailed(
    functions,
    clusters,
    expected,
):
    exact_keys = [
        "coeff_coords",
        "coeff_shape",
        "monomial_exponents",
        "variable_subsets",
        "variables",
    ]
    approx_keys = [
        "coeff_data",
    ]

    tex_keys = [
        "tex_coeff",
        "tex_prefactor",
    ]

    assert "functions" in expected
    assert len(expected["functions"]) == len(functions)
    assert len(expected["functions"]) == len(clusters)
    for i_orbit, orbit_expected in enumerate(expected["functions"]):
        assert len(functions[i_orbit]) == len(orbit_expected)
        assert len(clusters[i_orbit]) == len(orbit_expected)
        for i_equiv, equiv_expected in enumerate(orbit_expected):
            assert len(functions[i_orbit][i_equiv]) == len(equiv_expected)
            for i_func, func_expected in enumerate(equiv_expected):
                func = functions[i_orbit][i_equiv][i_func].to_dict()
                for key in exact_keys:
                    assert func_expected[key] == func[key]
                for key in approx_keys:
                    assert np.allclose(func_expected[key], func[key])
                for key in tex_keys:
                    if func_expected[key] == func[key]:
                        assert True
                    else:
                        if isinstance(func[key], list):
                            assert len(func_expected[key]) == len(func[key])
                            for i in range(len(func_expected[key])):
                                if func_expected[key][i] == func[key][i]:
                                    assert True
                                else:
                                    assert np.allclose(
                                        np.array(func_expected[key][i], dtype="float"),
                                        np.array(func[key][i], dtype="float"),
                                    )
                        else:
                            assert np.allclose(
                                np.array(func_expected[key], dtype="float"),
                                np.array(func[key], dtype="float"),
                            )


def make_discrete_magnetic_atom(
    name: str,
    value: Any,
    flavor: str = "C",
) -> xtal.Occupant:
    """Construct a discrete magnetic atomic occupant

    Parameters
    ----------
    name: str
        A "chemical name", which must be identical for atoms to be found symmetrically
        equivalent. The names are case sensitive, and “Va” is reserved for vacancies.
    value: Any
        The discrete value of the magnetic spin to associate with the constructed
        Occupant. If the type is `int` or `float` the value is converted to a
        size 1 array of float, other types are converted using `numpy.asarray`.
    flavor: str
        The magnetic spin "flavor", which must be one of varieties supported by CASM:
        `C`, `NC`, `SO`.

    Returns
    -------
    discrete_magnetic_atom: xtal.Occupant
        An :class:`~libcasm.xtal.Occupant` consisting of a single atom with the
        specified magnetic spin flavor and value.
    """
    if isinstance(value, (int, float)):
        value = np.array([value], dtype=np.float64)
    else:
        value = np.asarray(value, dtype=np.float64)

    return xtal.Occupant(
        name=name,
        atoms=[
            xtal.AtomComponent(
                name=name,
                coordinate=[0.0, 0.0, 0.0],
                properties={flavor + "magspin": value},
            )
        ],
    )



from ._basis_function_specs import (
    BasisFunctionSpecs,
)
from ._clex_basis_specs import (
    ClexBasisSpecs,
)
from ._cluster_functions import (
    ClusterFunctionsBuilder,
    make_local_point_functions,
    make_point_functions,
)
from ._discrete_functions import (
    get_occ_site_functions,
    make_chebychev_site_functions,
    make_composition_gram_matrix,
    make_composition_site_functions,
    make_direct_site_functions,
    make_occ_site_functions,
    make_occupation_site_functions,
    make_orthonormal_discrete_functions,
)
from ._matrix_rep import (
    ClusterMatrixRepBuilder,
    MakeVariableName,
    OrbitMatrixRepBuilder,
    make_cluster_dof_info,
    make_cluster_matrix_rep,
    make_cluster_permutation_rep,
    make_cluster_variables,
    make_equivalence_map_matrix_rep,
    make_equivalence_map_site_rep,
    make_global_dof_matrix_rep,
    make_global_variables,
)
from ._misc import (
    make_equivalents_generators,
    make_neighborhood,
    make_occevent_cluster_specs,
    make_symop_inverse,
    orbits_to_dict,
)



import copy
import math
import sys
from collections import namedtuple
from typing import Optional, TypeVar, Union

import numpy as np
import opt_einsum
import sparse

from casm.bset.misc import (
    factor_by_mode,
    irrational_to_tex_string,
)
from libcasm.counter import IntCounter


def is_canonical_coord(x: np.ndarray):
    """Check if a 1d array of tensor coordinates is sorted in decreasing order"""
    n = len(x) - 1
    for i in range(n):
        if x[i] < x[i + 1]:
            return False
    return True


def make_canonical_coord(x):
    """Return a copy of a 1d array of tensor coordinates, sorted in decreasing order"""
    return np.sort(x)[::-1]


def is_lowest_equivalent_coord(x: np.ndarray):
    """Check if a 1d array of tensor coordinates is sorted in ascending order"""
    n = len(x) - 1
    for i in range(n):
        if x[i] > x[i + 1]:
            return False
    return True


def make_lowest_equivalent_coord(x):
    """Return a copy of a 1d array of tensor coordinates, sorted in ascending order"""
    return np.sort(x)


def tensor_coord_to_monomial_exponents(x, n_size):
    """Convert tensor coordinates to vector of exponents form

    Example 1:

    .. code-block::

        # strain monomial {e_1}^4
        n_size=6
        x=np.array([0, 0, 0, 0], dtype='int')

        # equivalent vector of exponents
        n=np.array([4, 0, 0, 0, 0, 0], dtype='int')

    Example 2:

    .. code-block::

        # strain monomial {e_1}^2 {e_2}^1 {e_3}
        n_size=6
        x=np.array([2, 1, 0, 0], dtype='int')

        # equivalent vector of exponents
        n=np.array([2, 1, 1, 0, 0, 0], dtype='int')


    Parameters
    ----------
    x: np.ndarray[np.int]
        Tensor indices
    n_size: int
        Number of variable components, i.e. 6 for strain.

    Returns
    -------
    n: np.ndarray[np.int]
        Vector of monomial exponents
    """
    n = np.zeros((n_size,), dtype="int")
    for i in x:
        n[i] += 1
    return n


def monomial_exponents_to_tensor_coord(n):
    """Convert vectors of monomial exponents to tensor coordinates

    Parameters
    ----------
    n: np.ndarray[np.int]
        Vector of monomial exponents

    Returns
    -------
    x: np.ndarray[np.int]
        Tensor indices
    n_size: int
        Number of variable components, i.e. 6 for strain.
    """
    x = np.zeros((n.sum(),), dtype="int")
    j = 0
    i = n.size - 1
    while i > 0:
        while n[i] > 0:
            x[j] = i
            j += 1
            n[i] -= 1
        i -= 1
    return (x, n.size)


class FunctionRep:
    """Symmetry operation representation used to transform PolynomialFunction

    Currently, symmetry is applied to tensor coefficients using einstein summation
    notation. An improvement is planned to be made using the method described in
    Appendix C of J.C. Thomas, A. Van der Ven / Journal of the Mechanics and Physics of
    Solids 107 (2017) 76–95. DOI: 10.1016/j.jmps.2017.06.009

    """

    def __init__(
        self,
        matrix_rep: np.ndarray,
    ):
        """

        .. rubric:: Constructor

        Parameters
        ----------
        matrix_rep: np.ndarray[np.float64[m,m]]
            Describes the effect of applying a symmetry operation on a vector of
            variables.
        """

        self.matrix_rep = matrix_rep
        """np.ndarray[np.float64[m,m]]: Describes the effect of applying a symmetry 
        operation on a vector of variables.
        """

        self.sparse_matrix_rep = sparse.COO.from_numpy(matrix_rep)
        """sparse.COO: The `matrix_rep` as a sparse.COO array."""

    def __mul__(self, rhs):
        if isinstance(rhs, PolynomialFunction):
            dim = len(rhs.coeff.shape)
            result = rhs.copy()

            # A: result.coeff,
            # M: self.matrix_rep / self.sparse_matrix_rep
            #
            # Ex: indices_before, [i, j], indices_after
            #
            # dim = 1: A_a M_al -> A_l
            # i=0, j=1: [0], [0, 1], [1] # A_a M_al -> A_l
            #
            # dim = 2: A_ab M_al M_bm -> A_lm
            # i=0, j=2: [0, 1], [0, 2] -> [2, 1] # A_ab M_al -> A_lb
            # i=1, j=3: [2, 1], [1, 3] -> [2, 3] # A_lb M_bm -> A_lm
            #
            # dim = 3: A_abc M_al M_bm M_cn -> A_lmn
            # i=0, j=3: [0, 1, 2], [0, 3] -> [3, 1, 2] # A_abc M_al -> A_lbc
            # i=1, j=4: [3, 1, 2], [1, 4] -> [3, 4, 2] # A_lbc M_bm -> A_lmc
            # i=2, j=5: [3, 4, 2], [2, 5] -> [3, 4, 5] # A_lmc M_cn -> A_lmn
            #
            # etc.

            indices_before = [i for i in range(dim)]
            indices_after = [i for i in range(dim)]
            j = dim
            for i in range(dim):
                indices_after[i] = j
                result.coeff = opt_einsum.contract(
                    result.coeff,
                    indices_before,
                    self.sparse_matrix_rep,
                    [i, j],
                    indices_after,
                )
                indices_before = copy.deepcopy(indices_after)
                j += 1
            result.monomial_exponents = result.tensor_coords_to_monomial_exponents()

            result.make_canonical()
            result.prune()
            return result
        else:
            raise Exception(
                "Error in FunctionRep * X, with X="
                + str(type(rhs))
                + ": not implemented"
            )


VariableData = namedtuple(
    "VariableData",
    [
        "key",
        "cluster_site_index",
        "component_index",
        "site_basis_function_index",
        "neighborhood_site_index",
    ],
)


class Variable:
    r"""Represents a variable in a PolynomialFunction

    This class is used to keep the information necessary for collecting the degree
    of freedom (DoF) values necessary for evaluating polynomial functions.

    """

    def __init__(
        self,
        name: str,
        key: str,
        cluster_site_index: Optional[int] = None,
        component_index: Optional[int] = None,
        site_basis_function_index: Optional[int] = None,
        neighborhood_site_index: Optional[int] = None,
    ) -> object:
        R"""

        .. rubric:: Constructor

        Parameters
        ----------
        name: str
            Name used for identification and printing. For example:

            - "{e_1}", "{e_2}", etc. or "{E_{xx}}", "{E_{yy}}", etc. for strain
              degrees of freedom;
            - "{d_1}", "{d_2}", etc. or "dx", "dy", "dz" for displacement degrees of
              freedom;
            - "{\\\\\\phi_1}", "{\\\\\\phi_2}", etc. for occupation site basis
              functions, or "{\\\\\\phi_{a,1}}", "{\\\\\\phi_{a,2}}", etc. and
              "{\\\\\\phi_{b,1}}", "{\\\\\\phi_{b,2}}", etc. for occupation site basis
              functions on symmetrically distinct sites :math:`a` and :math:`b`.

        key: str
            Name of the degree of freedom (DoF) this variable represents.

        cluster_site_index: Optional[int] = None
            For site variables, the cluster site index of the site associated with the
            variable.

        component_index: Optional[int] = None
            For vector-valued continuous variables, the component of the vector this
            variable corresponds to.

        site_basis_function_index: Optional[int] = None
            For discrete occupation variables, the site basis function index
            this variable corresponds to.

        neighborhood_site_index: Optional[int] = None
            For site variables, the neighbor list index of the site associated with the
            variable.

            Used for printing expressions for evaluation of functions in terms of
            values on sites determined by the neighbor list position.

        """

        self.name = name
        R"""str: Name used for identification and printing.
        
        For example:
         
        - "{e_1}", "{e_2}", etc. or "{E_{xx}}", "{E_{yy}}", etc. for strain
          degrees of freedom; 
        - "{d_1}", "{d_2}", etc. or "dx", "dy", "dz" for displacement degrees of 
          freedom;
        - "{\phi_1}", "{\phi_2}", etc. for occupation site basis functions, or
          "{\phi_{a,1}}", "{\phi_{a,2}}", etc. and "{\phi_{b,1}}", "{\phi_{b,2}}", etc.
          for occupation site basis functions on symmetrically distinct sites :math:`a`
          and :math:`b`.
        """

        self.key = key
        """str: Name of the degree of freedom (DoF) this variable represents."""

        self.cluster_site_index = cluster_site_index
        """Optional[int]: For site variables, the cluster site index of the site 
        associated with the variable. 
        
        This is used for:

        - checking if a polynomial function includes all sites in a cluster
        - in combination with a cluster to determine the neighbor list index
        - printing latex formulas for functions on a cluster
        """

        self.component_index = component_index
        """Optional[int]: For vector-valued continuous variables, the component of \
        the vector this variable corresponds to.
        """

        self.site_basis_function_index = site_basis_function_index
        """Optional[int]: For occupation variables, the site basis function index \
        this variable corresponds to.
        """

        self.neighborhood_site_index = neighborhood_site_index
        """Optional[int]: For site variables, the neighbor list index of the site \
        associated with the variable.
        
        Used for printing expressions for evaluation of functions in terms of values on 
        sites determined by the neighbor list position.
        """

    def to_dict(self):
        data = {
            "name": self.name,
            "key": self.key,
        }
        if self.cluster_site_index is not None:
            data["cluster_site_index"] = self.cluster_site_index
        if self.component_index is not None:
            data["component_index"] = self.component_index
        if self.site_basis_function_index is not None:
            data["site_basis_function_index"] = self.site_basis_function_index
        if self.neighborhood_site_index is not None:
            data["neighborhood_site_index"] = self.neighborhood_site_index
        return data

    @staticmethod
    def from_dict(data):
        return Variable(
            name=data["name"],
            key=data["key"],
            cluster_site_index=data.get("cluster_site_index"),
            component_index=data.get("component_index"),
            site_basis_function_index=data.get("site_basis_function_index"),
            neighborhood_site_index=data.get("neighborhood_site_index"),
        )

    def _to_tuple(self):
        if self.neighborhood_site_index is None:
            return VariableData(
                self.key,
                self.cluster_site_index,
                self.component_index,
                self.site_basis_function_index,
                self.neighborhood_site_index,
            )
        else:
            # when neighborhood_site_index is set, we don't want to
            # compare cluster_site_index
            return VariableData(
                self.key,
                None,
                self.component_index,
                self.site_basis_function_index,
                self.neighborhood_site_index,
            )

    def __hash__(self):
        return hash(self._to_tuple())

    def __eq__(self, other):
        if isinstance(other, Variable):
            return self._to_tuple() == other._to_tuple()
        return NotImplemented


def is_subcluster_function(x: np.ndarray, variables: list[Variable], n_sites: int):
    site_included = [False] * n_sites
    for i in range(len(x)):
        csi = variables[x[i]].cluster_site_index
        if csi is not None:
            site_included[csi] = True
    return not all(site_included)


PolynomialFunctionType = TypeVar("PolynomialFunction")


class PolynomialFunction:
    r"""Represents a polynomial function as a sum monomial terms of the same order

    For example, if there are two variables, :math:`\vec{x} = [x_1, x_2]`,
    then PolynomialFunction can represent one of:

    .. math::

        f(\vec{x}) &= f_1 x{_1} + f_2 x{_2},

        f(\vec{x}) &= f_1 x{_1}^2 + f_2 x{_1} x{_2} + f_3 x{_2}^2,

        f(\vec{x}) &= f_1 x{_1}^3 + f_2 x{_1}^2 x{_2} + f_3 x{_1} x{_2}^2 + f_4 x{_2}^3,

        f(\vec{x}) &= f_1 x{_1}^4 + f_2 x{_1}^3 x{_2} + f_3 x{_1}^2 x{_2}^2 + \
        f_4 x{_1} x{_2}^3 + f_5 x{_2}^4,

        \mathrm{etc.}

    PolynomialFunction does not currently represent sums of monomials with
    different order (i.e. :math:`f_1 x{_1} + f_2 x{_1}^2`).

    """

    def __init__(
        self,
        coeff: sparse.COO,
        variables: list[Variable],
        variable_subsets: list[list[int]],
        tol: float = 1e-10,
    ):
        """

        .. rubric:: Constructor

        Parameters
        ----------
        coeff: sparse.COO
            A sparse tensor specifying non-zero monomial terms and corresponding
            coefficients.
        variables: list[Variable]
            The variables in the monomial terms.
        variable_subsets: list[list[int]]
            Lists of variables (as indices into the `variables` list) which mix under
            application of symmetry. For example, if all variables are the 6 strain
            variables, then `var_subsets=[[0,1,2,3,4,5]]`. If the variables are the 3
            displacements on each site in a cluster of 2 sites, then
            `variable_subsets=[[0, 1, 2], [3, 4, 5]]`.
        tol: float
            A tolerance used for pruning coefficients which are approximately zero.
        """

        self.coeff = coeff
        """sparse.COO: A sparse tensor holding coefficients of a polynomial.

        Example, polynomial with maximum order=2:

            A_ij = coeff,

        and the polynomial function is:

            f(x) = A_ij x_i x_j

        Example, a polynomial with maximum order=3:

            A_ijk = coeff,

        and the polynomial function is:

            f(x) = A_ijk x_i x_j x_k

        Operations (`*`, `+=`, `-=`, etc.) can be expected to also make
        `coeff` canonical and prune the coefficients that are
        approximately zero."""

        self.variables = variables
        """list[casm.bset.polynomial_functions.Variable]: The variables `x` of the 
        polynomial."""

        self.monomial_exponents = self.tensor_coords_to_monomial_exponents()
        """list[numpy.ndarray[numpy.int]]: The corresponding vectors of exponents for 
        the monomials (i.e. the non-zero tensor coefficients).

        For example:

        .. code-block:: Python

            n_size = len(variables)
            monomial_exponents[i] = tensor_coord_to_monomial_exponents(
                coeff.coords[:,i], n_size)

        The `monomial_exponents` is determined from `coeff` and must be updated
        whenever coeff.coords is modified."""

        self.variable_subsets = variable_subsets
        """list[list[int]]: Lists of variables (as indices into the `variables` list) 
        which mix under application of symmetry. 
        
        For example, if all variables are the 6 strain variables, then 
        `var_subsets=[[0,1,2,3,4,5]]`. If the variables are the 3 displacements on 
        each site in a cluster of 2 sites, then 
        `variable_subsets=[[0, 1, 2], [3, 4, 5]]`. This information is used in the 
        calculation of :func:`monomial_inner_product()`."""

        self.tol = tol
        """float: A tolerance used for pruning coefficients which are approximately 
        zero."""

    @staticmethod
    def zeros(
        shape: tuple[int],
        variables: list[Variable],
        variable_subsets: list[list[int]],
        tol: float = 1e-10,
    ):
        """Construct a new PolynomialFunction with all zero coefficients

        Parameters
        ----------
        shape: tuple[int]
            The shape of the coefficients tensor
        variables: list[Variable]
            The variables in the monomial terms.
        tol: float
            A tolerance used for pruning coefficients which are approximately zero.

        Returns
        -------
        function: PolynomialFunction
            A PolynomialFunction with all zero coefficients
        """
        coords = np.array([])
        data = np.array([])

        # Note: Constructing a zero-value sparse.COO gives a DeprecationWarning
        # in sparse 0.14.0, but it is erroneous and should not in the next release
        # (see https://github.com/pydata/sparse/pull/581)
        import warnings

        warnings.filterwarnings(
            "ignore",
            message="coords should be an ndarray. "
            "This will raise a ValueError in the future.",
        )

        return PolynomialFunction(
            coeff=sparse.COO(coords, data, shape=shape, prune=True),
            variables=variables,
            variable_subsets=variable_subsets,
            tol=tol,
        )

    def order(self):
        """Return the polynomial order"""
        return len(self.coeff.shape)

    def prune(self):
        """Update to remove sparse tensor components that are approximately zero."""
        mask = np.abs(self.coeff.data) > self.tol
        self.coeff = sparse.COO(
            self.coeff.coords[:, mask],
            self.coeff.data[mask],
            shape=self.coeff.shape,
            prune=True,
        )
        self.monomial_exponents = self.tensor_coords_to_monomial_exponents()
        return self

    def copy(self):
        return copy.deepcopy(self)

    def monomials(self):
        """Return individual monomials

        Returns
        -------
        monomials: list[PolynomialFunction]
            List of monomials which together sum to this polynomial
        """
        monomials = []
        for i in range(self.coeff.coords.shape[1]):
            mask = [i]

            monomials.append(
                PolynomialFunction(
                    coeff=sparse.COO(
                        self.coeff.coords[:, mask],
                        self.coeff.data[mask],
                        shape=self.coeff.shape,
                        prune=True,
                    ),
                    variables=self.variables,
                    variable_subsets=self.variable_subsets,
                    tol=self.tol,
                )
            )
        return monomials

    def __isub__(self, rhs: PolynomialFunctionType):
        self.coeff -= rhs.coeff
        self.make_canonical()
        self.prune()
        self.monomial_exponents = self.tensor_coords_to_monomial_exponents()
        return self

    def __iadd__(self, rhs: PolynomialFunctionType):
        self.coeff += rhs.coeff
        self.make_canonical()
        self.prune()
        self.monomial_exponents = self.tensor_coords_to_monomial_exponents()
        return self

    def __imul__(self, c: Union[float | Variable]):
        if isinstance(c, float):
            self.coeff *= c
            return self
        elif isinstance(c, Variable):
            i_var = self.variables.index(c)

            n_var = len(self.variables)
            order = self.coeff.coords.shape[0]
            n_monomial = self.coeff.coords.shape[1]
            coords_new = np.array(
                self.coeff.coords.copy().tolist() + [i_var] * n_monomial,
                dtype="int",
            )
            shape_new = [n_var] * (order + 1)

            self.coeff = sparse.COO(
                coords_new, self.coeff.data, shape=shape_new, prune=True
            )
            self.make_canonical()
            self.prune()
            self.monomial_exponents = self.tensor_coords_to_monomial_exponents()
            return self
        else:
            raise NotImplementedError

    def __mul__(self, c: Union[float | Variable]):
        result = self.copy()
        result *= c
        return result

    def __rmul__(self, c: Union[float | Variable]):
        return self * c

    def can_factor_by(self, var: Variable):
        """Check if a variable can be factored out of this polynomial

        Parameters
        ----------
        var: casm.bset.Variable
            Variable to be factored out.

        Returns
        -------
        result: bool
            True if this polynomial can be factored by `var`; False otherwise.
        """
        if var not in self.variables:
            return False
        i_var = self.variables.index(var)
        coords_list = self.coeff.coords.copy().transpose().tolist()
        for monomial_coords in coords_list:
            if i_var not in monomial_coords:
                return False
        return True

    def __itruediv__(self, c: Union[float | Variable]):
        if isinstance(c, float):
            self.coeff /= c
            return self
        elif isinstance(c, Variable):
            i_var = self.variables.index(c)

            n_var = len(self.variables)
            order = self.coeff.coords.shape[0]

            coords_list = self.coeff.coords.copy().transpose().tolist()
            for monomial_coords in coords_list:
                monomial_coords.remove(i_var)
            coords_new = np.array(coords_list, dtype="int").transpose()
            shape_new = [n_var] * (order - 1)

            self.coeff = sparse.COO(
                coords_new, self.coeff.data, shape=shape_new, prune=True
            )
            self.make_canonical()
            self.prune()
            self.monomial_exponents = self.tensor_coords_to_monomial_exponents()
            return self
        else:
            raise NotImplementedError

    def __truediv__(self, c: Union[float | Variable]):
        result = self.copy()
        result /= c
        return result

    def make_canonical(self):
        """Shift sparse tensor values from non-canonical coordinates to canonical \
        coordinates and prune zeros

        For example, coordinates `[0, 1]` and `[0, 2]` are non-canonical (increasing),
        so convert them from:

        .. code-block:: Python

            self.coeff.coords = [
                [0, 1, 2, 0],
                [1, 0, 0, 2],
            ]
            self.coeff.data = [1., 2., 3., 4.]

        To:

        .. code-block:: Python

            self.coeff.coords = [
                [1, 2],
                [0, 0],
            ]
            self.coeff.data = [3., 7.]

        Note that `self.coeff.data` must be updated appropriately.

        """
        if len(self.coeff.data) == 0:
            return self

        # copy canonical coords/data
        coords = self.coeff.coords.transpose().tolist()
        data = self.coeff.data.tolist()
        _coords = []
        _data = []
        for i in range(len(coords)):
            x = coords[i]
            v = data[i]
            if is_lowest_equivalent_coord(x):
                _coords.append(x)
                _data.append(v)
                continue
            x_canonical = make_lowest_equivalent_coord(x).tolist()

            found = False
            for j in range(len(_coords)):
                if _coords[j] == x_canonical:
                    found = True
                    _data[j] += v
                    break
            if not found:
                _coords.append(x_canonical)
                _data.append(v)

        # sort
        coords, data = zip(*sorted(zip(_coords, _data)))

        coords = np.array(coords).transpose()
        data = np.array(data)
        mask = np.abs(data) > self.tol
        self.coeff = sparse.COO(
            coords[:, mask],
            data[mask],
            shape=self.coeff.shape,
            prune=True,
        )
        self.monomial_exponents = self.tensor_coords_to_monomial_exponents()
        return self

    def tensor_coords_to_monomial_exponents(self):
        """Return sparse tensor coords as a list of vectors of monomial exponents

        Returns
        -------
        monomial_exponents: list[np.ndarray[np.int]]
            A list of vectors of monomial exponents, with one vector per monomial term
            with non-zero coefficient (i.e. one per coordinate in the `self.coeff`
            sparse tensor).
        """
        n_size = len(self.variables)
        coords = self.coeff.coords
        n_coords = coords.shape[1]
        return [
            tensor_coord_to_monomial_exponents(coords[:, i], n_size)
            for i in range(n_coords)
        ]

    def scalar_product(self, rhs: PolynomialFunctionType):
        r"""Scalar product

        Evaluates the Frobenius scalar product, :math:`\langle f, g \rangle_F`, of two
        polynomials, i.e. the sum of the :func:`monomial_inner_product` between each
        monomial term in the two polynomials. For example:

        .. math::

            f = f_1 {x_1}^2 + f_2 {x_1} {x_2} + f_3 {x_2}^2

            g = g_1 {x_1} * y + g_2 {x_2}^2

            \langle f, g \rangle_F = \\
                \langle f_1 {x_1}^2, g_1 {x_1} * {x_2} \rangle \\
                + \langle f_1 {x_1}^2, g_2 {x_2}^2 \rangle \\
                + \langle f_2 {x_1} y, g_1 {x_1} * {x_2} \rangle \\
                + \langle f_2 {x_1} y, g_2 {x_2}^2 \rangle \\
                + \langle f_3 {x_2}^2, g_1 {x_1} * {x_2} \rangle \\
                + \langle f_3 {x_2}^2, g_2 {x_2}^2 \rangle

        The :func:`monomial_inner_product` is defined to be 0 for monomials differing
        only by their exponents, so this simplifies to:

        .. math::

            \langle f, g \rangle_F = \\
                \langle f_2 {x_1} y, g_1 {x_1} * {x_2} \rangle \\
                + \langle f_3 {x_2}^2, g_2 {x_2}^2 \rangle


        Parameters
        ----------
        rhs: PolynomialFunction
            The second polynomial.

        Returns
        -------
        product: float
            Frobeinius scalar product of this PolynomialFunction and `rhs`, as defined
            above.
        """
        a_exp = self.monomial_exponents
        a_coeff = self.coeff.data
        a_size = len(a_exp)

        b_exp = rhs.monomial_exponents
        b_coeff = rhs.coeff.data
        b_size = len(b_exp)

        prod = 0.0
        for i_a in range(a_size):
            for i_b in range(b_size):
                prod += monomial_inner_product(
                    a_coeff[i_a],
                    a_exp[i_a],
                    b_coeff[i_b],
                    b_exp[i_b],
                    self.variable_subsets,
                )
        return prod

    def normalize(self):
        self.coeff /= np.sqrt(self.scalar_product(self))
        return self

    def __lt__(self, rhs):
        a = self.coeff.coords
        b = rhs.coeff.coords
        if a.shape[1] != b.shape[1]:
            return a.shape[1] < b.shape[1]
        elif not np.all(a == b):
            return a.T.tolist() < b.T.tolist()
        a = self.coeff.data
        b = self.coeff.data

        if np.allclose(a, b):
            return False
        return a.tolist() < b.tolist()

    def __gt__(self, rhs):
        return rhs < self

    def __le__(self, rhs):
        return not rhs < self

    def __ge__(self, rhs):
        return not self < rhs

    def __eq__(self, rhs):
        return not self < rhs and not rhs < self

    def __ne__(self, rhs):
        return self < rhs or rhs < self

    def _basic_print(self):
        """Basic printing for development"""
        limit = len(self.variables) ** 2
        max_pow = 2

        a_tex = 0
        coeff_tex = []

        if len(self.coeff.data) != 0:
            a, factored_data = factor_by_mode(self.coeff.data)

            a_tex = irrational_to_tex_string(
                a, limit=limit, max_pow=max_pow, abs_tol=1e-5
            )
            coeff_tex = [
                irrational_to_tex_string(x, limit=limit, max_pow=max_pow, abs_tol=1e-5)
                for x in factored_data
            ]

        print("coordinates:")
        print(self.coeff.coords)
        print("coefficients:")
        print(f"{a_tex} * [", end="")
        for s in coeff_tex:
            print(f"{s}, ", end="")
        print("]")

    def to_dict(self):
        """Represent as a Python dict"""
        limit = len(self.variables) ** 2
        max_pow = 2

        a_tex = 0
        coeff_tex = []

        if len(self.coeff.data) != 0:
            a, factored_data = factor_by_mode(self.coeff.data)

            a_tex = irrational_to_tex_string(
                a, limit=limit, max_pow=max_pow, abs_tol=1e-5
            )
            coeff_tex = [
                irrational_to_tex_string(x, limit=limit, max_pow=max_pow, abs_tol=1e-5)
                for x in factored_data
            ]

        tex_prefactor_str = str(a_tex)
        tex_coeff_str = [str(x) for x in coeff_tex]

        return {
            "coeff_coords": self.coeff.coords.T.tolist(),
            "coeff_data": self.coeff.data.tolist(),
            "coeff_shape": list(self.coeff.shape),
            "variables": [var.to_dict() for var in self.variables],
            "variable_subsets": self.variable_subsets,
            "monomial_exponents": [x.tolist() for x in self.monomial_exponents],
            "tex_prefactor": tex_prefactor_str,
            "tex_coeff": tex_coeff_str,
            "tol": self.tol,
        }

    @staticmethod
    def from_dict(data):
        """Construct from a Python dict"""
        variables = [Variable.from_dict(x) for x in data["variables"]]
        variable_subsets = data["variable_subsets"]
        coords = np.array(data["coeff_coords"], dtype=int).T
        _data = np.array(data["coeff_data"])
        shape = tuple(data["coeff_shape"])
        tol = data.get("tol", 1e-10)
        return PolynomialFunction(
            coeff=sparse.COO(coords, _data, shape=shape, prune=True),
            variables=variables,
            variable_subsets=variable_subsets,
            tol=tol,
        )

    def latex_formula(
        self,
    ):
        """Latex printing for development
        It might be good to take variables as strings."""
        variables = self.variables
        exponents = self.monomial_exponents
        a, factored_data = factor_by_mode(self.coeff.data)

        limit = len(variables) ** 2
        max_pow = 2

        # build terms
        latex_formula = ""
        for i, val in enumerate(
            factored_data
        ):  # len(factored_data) is the number of terms
            coeff = a * factored_data[i]
            if np.isclose(coeff, 1):
                coeff_tex = ""
            else:
                coeff_tex = irrational_to_tex_string(
                    a * factored_data[i], limit=limit, max_pow=max_pow, abs_tol=1e-5
                )
            polynomial_tex = ""
            for j, exp in enumerate(exponents[i]):
                if exp == 0:
                    continue  # constant contribution
                if exp == 1:
                    polynomial_tex += f" * ({variables[j].name})"
                else:
                    polynomial_tex += f" * ({variables[j].name})^{exp}"
            polynomial_tex = polynomial_tex.strip("* ")
            latex_formula += f"{coeff_tex} * [{polynomial_tex}] + "

        return latex_formula.strip("*+ ")

    def _latex_print(self):
        """Latex printing for development
        It might be good to take variables as strings."""
        print(self.latex_formula())


def monomial_inner_product(
    a_coeff: float,
    n_a: list[int],
    b_coeff: float,
    n_b: list[int],
    variable_subsets: list[list[int]],
) -> float:
    r"""Evaluates the monomial inner product

    Evaluates the monomial inner product as defined by Eq. (C.4) of :cite:`THOMAS2017a`:

    .. math::

        \left \langle \alpha^{(\vec{n})}\prod_{m} x_m^{n_m},
        \beta^{(\vec{n}')}\prod_{m'} x_{m'}^{n'_{m'}} \right \rangle
        = \alpha^{(\vec{n})}\beta^{(\vec{n}')}
        \frac{1}{\prod_l(\sum_{m \in S_l} n_{m})!}
        \prod_m\delta_{n_{m}n'_{m}}n_{m}!

    where:

    - :math:`\vec{x}`: the vector of variable components,
      i.e. for strain, :math:`\vec{x}=[e_1, e_2, e_3, e_4, e_5, e_6 ]`.
    - :math:`\alpha^{(\vec{n})}`: the coefficient for the first monomial
    - :math:`\vec{n}`: the vector of exponents for the first monomial
    - :math:`\beta^{(\vec{n}')}`: the coefficient for the second monomial
    - :math:`\vec{n}'`: the vector of exponents for the second monomial
    - :math:`S_l`: the subsets of indices of variables which mix under application of
      symmetry. For example, in a cluster of 2 sites with 3 site displacement degrees
      of freedom on each site (:math:`dx`, :math:`dy`, :math:`dz`), there are 6 total
      variables, the first three (with indices [0,1,2]) being the displacements on the
      first site and the second three (with indices [3, 4, 5]) being the displacements
      on the second site. Then the variable subsets would be :math:`S_1=[0, 1, 2]` and
      :math:`S_2=[3, 4, 5]`.

    Parameters
    ----------
    a_coeff: float
        The coefficient for the first monomial, :math:`\alpha^{(\vec{n})}`.
    n_a: np.ndarray[np.int]
        The vector of exponents for the first monomial, :math:`\vec{n}`
    b_coeff: float
        The coefficient for the first monomial, :math:`\beta^{(\vec{n})}`.
    n_b: np.ndarray[np.int]
        The vector of exponents for the second monomial, :math:`\vec{n}'`
    variable_subsets: list[list[int]]
        The subsets of variables which mix under application of symmetry, as indices
        into the vectors of exponents.

    Returns
    -------
    product: float
        The monomial inner product, as defined above.
    """
    if n_a.size != n_b.size:
        raise Exception("Error in monomial_inner_product: size mismatch")

    # Evaluate: factorial_prod = \prod_m\delta_{n_{m}n'_{m}}n_{m}!
    factorial_prod = 1
    for m in range(n_a.size):
        # if any delta_{n_{m}n'_{m}} == 0, can return 0.0
        if n_a[m] != n_b[m]:
            return 0.0
        factorial_prod *= math.factorial(n_a[m])

    factorial_quotient = 1.0
    for subset in variable_subsets:
        sum = 0
        for m in subset:
            sum += n_a[m]
        factorial_quotient *= math.factorial(sum)

    # return a_coeff * b_coeff * (1.0 / math.factorial(np.sum(n_a))) * factorial_prod
    return a_coeff * b_coeff * factorial_prod / factorial_quotient


def gram_schmidt(
    functions: list[PolynomialFunction],
) -> list[PolynomialFunction]:
    """Apply the Gram-Schmidt procedure to orthonormalize a list of PolynomialFunction

    Parameters
    ----------
    functions: list[PolynomialFunction]
        A list of :class:`~casm.bset.coo.PolynomialFunction`.

    Returns
    -------
    orthonormalized_functions: list[PolynomialFunction]
        Functions after applying Gram-Schmidt
    """
    orthonormalized_functions = []
    for f in functions:
        if len(f.coeff.data) == 0:
            continue
        next = f.copy()
        next.make_canonical()
        next.prune()
        for g in orthonormalized_functions:
            next -= next.scalar_product(g) * g
        if len(next.coeff.data) != 0:
            next.normalize()
            orthonormalized_functions.append(next)
    orthonormalized_functions.sort()
    return orthonormalized_functions


class ExponentSumConstraint:
    """Data structure that specifies an exponent sum for filtering monomials"""

    def __init__(
        self,
        variables: list[int],
        sum: list[int],
    ):
        """

        .. rubric:: Constructor

        Parameters
        ----------
        variables: list[int]
            Indices of the variables included in the constraint
        sum: list[int]
            The sum of the exponents of the specified variables
            must be in the `sum` list for the constraint to
            be satisfied.
        """
        self.variables = variables
        """list[int]: Indices of the variables in the
        :py:data:`PolynomialFunction.variables <casm.bset.polynomial_functions.PolynomialFunction.variables>`
        list that are included in the constraint."""  # noqa

        self.sum = sum
        """list[int]: The sum of the exponents of the specified variables must be in \
        the `sum` list for the constraint to be satisfied."""

    def satisfied(self, monomial_exponents):
        curr_sum = 0
        for i_var in self.variables:
            curr_sum += monomial_exponents[i_var]
        return curr_sum in self.sum

    def to_dict(self):
        """Represent as a Python dict"""
        return {"variables": self.variables, "sum": self.sum}

    @staticmethod
    def from_dict(data):
        """Construct from a Python dict"""
        return ExponentSumConstraint(
            variables=data["variables"],
            sum=data["sum"],
        )


def all_constraints_satisfied(constraints, monomial_exponents):
    for constraint in constraints:
        if not constraint.satisfied(monomial_exponents):
            return False
    return True


def make_symmetry_adapted_polynomials(
    matrix_rep: list[np.ndarray],
    variables: list[Variable],
    variable_subsets: list[list[int]],
    max_poly_order: int,
    min_poly_order: int = 1,
    orthonormalize_in_place: bool = True,
    constraints: list[ExponentSumConstraint] = [],
    eps: float = 1e-14,
    verbose: bool = False,
) -> list[PolynomialFunction]:
    """Generate symmetry adapted polynomial functions

    This method systematically enumerates monomial terms of varying order, applies
    group operations to construct symmetry adapted polynomials, and then orthogonalizes
    the results using the Gram-Schmidt method.

    Notes
    -----

    To support cluster expansion function construction, if any variable has a
    `cluster_site_index` which is not None, then the set of
    `cluster_site_index` is found, and monomials are checked to ensure that they
    have at least one variable on every site in the cluster. Any monomial which
    only has variables on a sub-cluster of sites is skipped.

    Parameters
    ----------
    matrix_rep: list[np.ndarray[np.float]]
        Matrix representation for the symmetry groups acting on the variables.
    variables: list[Variable]
        Describes the variable elements of the vector that `matrix_rep` acts on.
    variable_subsets: list[list[int]]
        The indices of Variable in `variables` which mix under application of symmetry,
        or permute as a group. A subset could be strain variables, displacement
        variables on a site, or occupant site basis functions on a site.
    max_poly_order: int
        Maximum order (sum of exponents) in monomials of the generated
        polynomials.
    min_poly_order: int = 1
        Minimum order (sum of exponents) in monomials of the generated
        polynomials.
    orthonormalize_in_place = True
        If True, orthonormalize symmetry adapted polynomials as they are generated,
        otherwise generate all symmetry adapted polynomials of a particular order and
        then orthonormalize.
    constraints: list[ExponentSumConstraint]
        If any constraint is not satisfied, the candidate monomial is skipped. This
        can be used to ensure one and only one of mutually exclusive discrete variables
        is included in any monomial.
    eps: float = 1e-14
        Tolerance used for identifying zeros in the matrix representations.
    verbose: bool = False
        Print progress statements

    Returns
    -------
    functions: list[PolynomialFunction]
        A list of :class:`~casm.bset.polynomial_function.PolynomialFunction` that are
        invariant to symmetry operations in `matrix_rep` and have been orthonormalized
        according to the scalar product defined in :cite:`THOMAS2017a`.
    """

    n_variables = len(variables)

    function_rep = []
    for i in range(len(matrix_rep)):
        M = matrix_rep[i].copy()
        M[np.abs(M) < eps] = 0.0
        S = FunctionRep(matrix_rep=M)
        function_rep.append(S)

    # check if cluster function:
    is_cluster_function = False
    n_sites = None
    for var in variables:
        if var.cluster_site_index is not None:
            is_cluster_function = True
            if n_sites is None or var.cluster_site_index >= n_sites:
                n_sites = var.cluster_site_index + 1

    # Holds orthonormalized symmetry adapted polynomials of all orders
    all_functions = []

    in_place = orthonormalize_in_place

    for poly_order in range(min_poly_order, max_poly_order + 1):
        if verbose:
            print(f"Working on symmetry adapted polynomials of order {poly_order}")
            if in_place:
                print("Progress (generated / orthonormalized):")
            else:
                print("Progress (generated):")
            sys.stdout.flush()

        # Holds symmetry adapted polynomials of order poly_order
        functions = []
        if in_place:
            orthonormalized_functions = []

        # Counter over all sparse tensor coordinates
        counter = IntCounter(
            initial=[0] * poly_order,
            final=[n_variables - 1] * poly_order,
            increment=[1] * poly_order,
        )

        shape = [n_variables] * poly_order

        f_count = 0
        for x in counter:
            # Using the sparse tensor representation, permutations of coords
            # indicate an equivalent monomial so we skip all but the canonical coord
            if not is_canonical_coord(x):
                continue

            # For cluster functions, want at least one dof from each site:
            if is_cluster_function and is_subcluster_function(x, variables, n_sites):
                continue

            # Skip monomials that do not satisfy exponent constraints
            monomial_exponents = tensor_coord_to_monomial_exponents(x, n_variables)
            if not all_constraints_satisfied(constraints, monomial_exponents):
                continue

            # print("Adding monomial:", x)

            # Create polynomial function coefficients tensor
            coords = x.reshape((-1, 1))  # 2d array with single column containing x
            data = [1.0]  # coefficient value for monomial specified by x
            f_init = PolynomialFunction(
                coeff=sparse.COO(coords, data, shape=shape, prune=True),
                variables=variables,
                variable_subsets=variable_subsets,
            )

            # Apply Reynolds operator to construct symmetrized polynomials
            f_sum = PolynomialFunction.zeros(
                shape=shape,
                variables=variables,
                variable_subsets=variable_subsets,
            )
            for i, S in enumerate(function_rep):
                f_sum += S * f_init

            # If function is not enforced to be 0 by symmetry,
            # then add it to the growing list of symmetry adapted functions
            if len(f_sum.coeff.data) != 0:
                if in_place:
                    next = f_sum.copy()
                    for g in orthonormalized_functions:
                        next -= next.scalar_product(g) * g
                    if len(next.coeff.data) != 0:
                        next.normalize()
                        orthonormalized_functions.append(next)
                functions.append(f_sum)

            if verbose:
                print(".", end="")
                f_count += 1
                sys.stdout.flush()
                if f_count % 10 == 0:
                    if in_place:
                        print(f" {len(functions)} / {len(orthonormalized_functions)}")
                    else:
                        print(f" {len(functions)}")

        if verbose:
            print()
            print(f"- Checked {f_count} monomials of order {poly_order}")
            print(
                f"- Generated {len(functions)} symmetry "
                f"adapted polynomials of order {poly_order}"
            )

        if not in_place:
            orthonormalized_functions = gram_schmidt(functions)
        else:
            orthonormalized_functions.sort()
        all_functions += orthonormalized_functions

        if verbose:
            print(
                f"- Generated {len(orthonormalized_functions)} "
                f"orthonormalized functions of order {poly_order}"
            )
            print(f"- Generated {len(all_functions)} total orthonormalized functions")
            print("")

    return all_functions



import json

import numpy as np
from utils.helpers import (
    assert_expected_cluster_functions_detailed,
    make_discrete_magnetic_atom,
)

import libcasm.configuration as casmconfig
import libcasm.xtal as xtal
from casm.bset import (
    build_cluster_functions,
)
from casm.bset.cluster_functions import (
    get_occ_site_functions,
)


def test_occ_fcc_1a(session_shared_datadir):
    """Baseline test - occupation basis functions

    To be compared with test_occ_fcc_1b,
    where the occupants on i_sublat=2 are ordered differently.
    """

    xtal_prim = xtal.Prim(
        lattice=xtal.Lattice(
            column_vector_matrix=np.eye(3),
        ),
        coordinate_frac=np.array(
            [
                [0.0, 0.0, 0.0],
                [0.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 0.5, 0.0],
            ]
        ).T,
        occ_dof=[
            ["A", "B"],
            ["B", "C"],
            ["B", "C"],
            ["B", "C"],
        ],
    )
    # print(xtal.pretty_json(xtal_prim.to_dict()))

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {"occ": {"site_basis_functions": "occupation"}}
            },
        },
        verbose=False,
    )
    functions, clusters = (builder.functions, builder.clusters)

    for b in [0, 1, 2, 3]:
        expected = np.array(
            [
                [1.0, 1.0],
                [0.0, 1.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    # import os
    # import pathlib
    # from utils.helpers import print_expected_cluster_functions_detailed
    #
    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_site_functions_fcc_1.json",
    # )
    with open(session_shared_datadir / "expected_occ_site_functions_fcc_1.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))


def test_occ_fcc_1b(session_shared_datadir):
    """Change order of occupants on i_sublat=2

    This affects the site basis function (phi) values on that sublattice,
    but the generated functions are the same
    (because they are functions of the phi which remain in the same order).
    """
    xtal_prim = xtal.Prim(
        lattice=xtal.Lattice(
            column_vector_matrix=np.eye(3),
        ),
        coordinate_frac=np.array(
            [
                [0.0, 0.0, 0.0],
                [0.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 0.5, 0.0],
            ]
        ).T,
        occ_dof=[
            ["A", "B"],
            ["B", "C"],
            ["C", "B"],
            ["B", "C"],
        ],
    )
    # print(xtal.pretty_json(xtal_prim.to_dict()))

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {"occ": {"site_basis_functions": "occupation"}}
            },
        },
        verbose=False,
    )
    functions, clusters = (builder.functions, builder.clusters)

    for b in [0, 1, 3]:
        expected = np.array(
            [
                [1.0, 1.0],
                [0.0, 1.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    for b in [2]:
        expected = np.array(
            [
                [1.0, 1.0],
                [1.0, 0.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    ## Compare to functions generated in test 1a ##

    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_site_functions_fcc_1b.json",
    # )
    with open(session_shared_datadir / "expected_occ_site_functions_fcc_1.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))


def test_occ_fcc_2a(session_shared_datadir):
    """Baseline test - ternary occupation basis functions

    To be compared with test_occ_fcc_2b,
    where the occupants on i_sublat=2 are ordered differently.
    """

    xtal_prim = xtal.Prim(
        lattice=xtal.Lattice(
            column_vector_matrix=np.eye(3),
        ),
        coordinate_frac=np.array(
            [
                [0.0, 0.0, 0.0],
                [0.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 0.5, 0.0],
            ]
        ).T,
        occ_dof=[
            ["A", "B"],
            ["B", "C", "D"],
            ["B", "C", "D"],
            ["B", "C", "D"],
        ],
    )
    # print(xtal.pretty_json(xtal_prim.to_dict()))

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {"occ": {"site_basis_functions": "occupation"}}
            },
        },
        verbose=False,
    )
    functions, clusters = (builder.functions, builder.clusters)

    for b in [0]:
        expected = np.array(
            [
                [1.0, 1.0],
                [0.0, 1.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    for b in [1, 2, 3]:
        expected = np.array(
            [
                [1.0, 1.0, 1.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    # import os
    # import pathlib
    # from utils.helpers import print_expected_cluster_functions_detailed
    #
    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_site_functions_fcc_2.json",
    # )
    with open(session_shared_datadir / "expected_occ_site_functions_fcc_2.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))


def test_occ_fcc_2b(session_shared_datadir):
    """Change order of occupants

    This affects the site basis function (phi) values on that sublattice,
    but the generated functions are the same as test_occ_fcc_2a
    (because they are functions of the phi which remain in the same order).
    """

    occ_dof = [
        ["A", "B"],
        ["B", "C", "D"],
        ["B", "D", "C"],
        ["C", "D", "B"],
    ]
    xtal_prim = xtal.Prim(
        lattice=xtal.Lattice(
            column_vector_matrix=np.eye(3),
        ),
        coordinate_frac=np.array(
            [
                [0.0, 0.0, 0.0],
                [0.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 0.5, 0.0],
            ]
        ).T,
        occ_dof=occ_dof,
    )
    prim = casmconfig.Prim(xtal_prim)
    occ_symgroup_rep = prim.occ_symgroup_rep

    for i_factor_group, occ_op_rep in enumerate(occ_symgroup_rep):
        site_rep = prim.integral_site_coordinate_symgroup_rep[i_factor_group]
        for i_sublat_before, occ_sublat_rep in enumerate(occ_op_rep):
            site_before = xtal.IntegralSiteCoordinate(i_sublat_before, [0, 0, 0])
            site_after = site_rep * site_before
            i_sublat_after = site_after.sublattice()
            for i_occ_before in range(len(occ_sublat_rep)):
                i_occ_after = occ_sublat_rep[i_occ_before]
                assert (
                    occ_dof[i_sublat_before][i_occ_before]
                    == occ_dof[i_sublat_after][i_occ_after]
                )

    # print(xtal.pretty_json(xtal_prim.to_dict()))

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {"occ": {"site_basis_functions": "occupation"}}
            },
        },
        verbose=False,
    )
    functions, clusters = (builder.functions, builder.clusters)

    for b in [0]:
        expected = np.array(
            [
                [1.0, 1.0],
                [0.0, 1.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    for b in [1]:
        # B, C, D
        expected = np.array(
            [
                [1.0, 1.0, 1.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    for b in [2]:
        # B, D, C
        expected = np.array(
            [
                [1.0, 1.0, 1.0],
                [0.0, 0.0, 1.0],
                [0.0, 1.0, 0.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    for b in [3]:
        # C, D, B
        expected = np.array(
            [
                [1.0, 1.0, 1.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_site_functions_fcc_2.json",
    # )
    with open(session_shared_datadir / "expected_occ_site_functions_fcc_2.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))


def test_occ_fcc_3a(session_shared_datadir):
    """Baseline - prim with occupation DoF with chemistry and magspin

    To be compared with test_occ_fcc_3b,
    where the occupants are ordered differently.
    """

    # Lattice vectors
    lattice = xtal.Lattice(np.eye(3))

    # Basis sites positions, as columns of a matrix,
    # in fractional coordinates with respect to the lattice vectors
    coordinate_frac = np.array(
        [
            [0.0, 0.0, 0.0],
            [0.0, 0.5, 0.5],
            [0.5, 0.0, 0.5],
            [0.5, 0.5, 0.0],
        ]
    ).transpose()

    # Occupation degrees of freedom (DoF)
    occupants = {
        "A.up": make_discrete_magnetic_atom(name="A", value=1, flavor="C"),
        "A.down": make_discrete_magnetic_atom(name="A", value=-1, flavor="C"),
        "B.up": make_discrete_magnetic_atom(name="B", value=1, flavor="C"),
        "B.down": make_discrete_magnetic_atom(name="B", value=-1, flavor="C"),
        "C": xtal.make_atom(name="C"),
    }
    occ_dof = [
        ["C", "A.up", "A.down"],
        ["C", "A.up", "A.down", "B.up", "B.down"],
        ["C", "A.up", "A.down", "B.up", "B.down"],
        ["C", "A.up", "A.down", "B.up", "B.down"],
    ]

    xtal_prim = xtal.Prim(
        lattice=lattice,
        coordinate_frac=coordinate_frac,
        occ_dof=occ_dof,
        occupants=occupants,
    )
    prim = casmconfig.Prim(xtal_prim)
    assert len(prim.factor_group.elements) == 96
    occ_symgroup_rep = prim.occ_symgroup_rep

    for i_factor_group, occ_op_rep in enumerate(occ_symgroup_rep):
        site_rep = prim.integral_site_coordinate_symgroup_rep[i_factor_group]
        for i_sublat_before, occ_sublat_rep in enumerate(occ_op_rep):
            site_before = xtal.IntegralSiteCoordinate(i_sublat_before, [0, 0, 0])
            site_after = site_rep * site_before
            i_sublat_after = site_after.sublattice()
            for i_occ_before in range(len(occ_sublat_rep)):
                i_occ_after = occ_sublat_rep[i_occ_before]

                orientation_name_before = occ_dof[i_sublat_before][i_occ_before]
                orientation_name_after = occ_dof[i_sublat_after][i_occ_after]

                # assert occupants map (chemical name match)
                assert (
                    occupants[orientation_name_before].name()
                    == occupants[orientation_name_after].name()
                )

    # print(xtal.pretty_json(xtal_prim.to_dict()))

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {
                    "occ": {
                        "site_basis_functions": {
                            "type": "occupation",
                            "reference_occ": ["C", "C", "C", "C"],
                        }
                    }
                }
            },
        },
        verbose=False,
    )
    functions, clusters = (builder.functions, builder.clusters)

    for b in [0]:
        # "C", "A.up", "A.down"
        expected = np.array(
            [
                [1.0, 1.0, 1.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    for b in [1, 2, 3]:
        # "C", "A.up", "A.down", "B.up", "B.down"
        expected = np.array(
            [
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [0.0, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_site_functions_fcc_3.json",
    # )
    with open(session_shared_datadir / "expected_occ_site_functions_fcc_3.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))


def test_occ_fcc_3b(session_shared_datadir):
    """Change order of occupants

    This affects the site basis function (phi) values on that sublattice,
    but the generated functions are the same as test_occ_fcc_3a
    (because they are functions of the phi which remain in the same order).
    """

    # Lattice vectors
    lattice = xtal.Lattice(np.eye(3))

    # Basis sites positions, as columns of a matrix,
    # in fractional coordinates with respect to the lattice vectors
    coordinate_frac = np.array(
        [
            [0.0, 0.0, 0.0],
            [0.0, 0.5, 0.5],
            [0.5, 0.0, 0.5],
            [0.5, 0.5, 0.0],
        ]
    ).transpose()

    # Occupation degrees of freedom (DoF)
    occupants = {
        "A.up": make_discrete_magnetic_atom(name="A", value=1, flavor="C"),
        "A.down": make_discrete_magnetic_atom(name="A", value=-1, flavor="C"),
        "B.up": make_discrete_magnetic_atom(name="B", value=1, flavor="C"),
        "B.down": make_discrete_magnetic_atom(name="B", value=-1, flavor="C"),
        "C": xtal.make_atom(name="C"),
    }
    occ_dof = [
        ["A.up", "A.down", "C"],
        ["A.up", "A.down", "B.up", "B.down", "C"],
        ["A.up", "A.down", "B.up", "B.down", "C"],
        ["A.up", "A.down", "B.up", "B.down", "C"],
    ]

    xtal_prim = xtal.Prim(
        lattice=lattice,
        coordinate_frac=coordinate_frac,
        occ_dof=occ_dof,
        occupants=occupants,
    )
    prim = casmconfig.Prim(xtal_prim)
    assert len(prim.factor_group.elements) == 96
    occ_symgroup_rep = prim.occ_symgroup_rep

    for i_factor_group, occ_op_rep in enumerate(occ_symgroup_rep):
        site_rep = prim.integral_site_coordinate_symgroup_rep[i_factor_group]
        for i_sublat_before, occ_sublat_rep in enumerate(occ_op_rep):
            site_before = xtal.IntegralSiteCoordinate(i_sublat_before, [0, 0, 0])
            site_after = site_rep * site_before
            i_sublat_after = site_after.sublattice()
            for i_occ_before in range(len(occ_sublat_rep)):
                i_occ_after = occ_sublat_rep[i_occ_before]

                orientation_name_before = occ_dof[i_sublat_before][i_occ_before]
                orientation_name_after = occ_dof[i_sublat_after][i_occ_after]

                # assert occupants map (chemical name match)
                assert (
                    occupants[orientation_name_before].name()
                    == occupants[orientation_name_after].name()
                )

    # print(xtal.pretty_json(xtal_prim.to_dict()))

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {
                    "occ": {
                        "site_basis_functions": {
                            "type": "occupation",
                            "reference_occ": ["C", "C", "C", "C"],
                        }
                    }
                }
            },
        },
        verbose=False,
    )
    functions, clusters = (builder.functions, builder.clusters)

    for b in [0]:
        # "A.up", "A.down", "C"
        expected = np.array(
            [
                [1.0, 1.0, 1.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    for b in [1, 2, 3]:
        # "A.up", "A.down", "B.up", "B.down", "C"
        expected = np.array(
            [
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0, 0.0],
            ]
        )
        assert np.allclose(builder.occ_site_functions[b]["value"], expected)

    # import os
    # import pathlib
    # from utils.helpers import print_expected_cluster_functions_detailed
    #
    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_site_functions_fcc_3.json",
    # )
    with open(session_shared_datadir / "expected_occ_site_functions_fcc_3.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))


def test_occ_fcc_4a(session_shared_datadir):
    """Baseline - prim with occupation DoF with molecular orientation

    To be compared with test_occ_fcc_4b,
    where the occupants are ordered differently.
    """

    mol_x = xtal.Occupant(
        name="mol",
        atoms=[
            xtal.AtomComponent(name="B", coordinate=[-0.1, 0.0, 0.0], properties={}),
            xtal.AtomComponent(name="B", coordinate=[0.1, 0.0, 0.0], properties={}),
        ],
    )
    mol_y = xtal.Occupant(
        name="mol",
        atoms=[
            xtal.AtomComponent(name="B", coordinate=[0.0, -0.1, 0.0], properties={}),
            xtal.AtomComponent(name="B", coordinate=[0.0, 0.1, 0.0], properties={}),
        ],
    )
    mol_z = xtal.Occupant(
        name="mol",
        atoms=[
            xtal.AtomComponent(name="B", coordinate=[0.0, 0.0, -0.1], properties={}),
            xtal.AtomComponent(name="B", coordinate=[0.0, 0.0, 0.1], properties={}),
        ],
    )
    atom_A = xtal.Occupant(
        name="A",
        atoms=[
            xtal.AtomComponent(name="A", coordinate=[0.0, 0.0, 0.0], properties={}),
        ],
    )
    occupants = {"mol.x": mol_x, "mol.y": mol_y, "mol.z": mol_z, "A": atom_A}
    occ_dof = [
        ["A"],
        ["A", "mol.x", "mol.y", "mol.z"],
        ["A", "mol.x", "mol.y", "mol.z"],
        ["A", "mol.x", "mol.y", "mol.z"],
    ]
    xtal_prim = xtal.Prim(
        lattice=xtal.Lattice(
            column_vector_matrix=np.eye(3),
        ),
        coordinate_frac=np.array(
            [
                [0.0, 0.0, 0.0],
                [0.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 0.5, 0.0],
            ]
        ).T,
        occ_dof=occ_dof,
        occupants=occupants,
    )
    prim = casmconfig.Prim(xtal_prim)
    assert len(prim.factor_group.elements) == 48
    occ_symgroup_rep = prim.occ_symgroup_rep

    for i_factor_group, occ_op_rep in enumerate(occ_symgroup_rep):
        site_rep = prim.integral_site_coordinate_symgroup_rep[i_factor_group]
        for i_sublat_before, occ_sublat_rep in enumerate(occ_op_rep):
            site_before = xtal.IntegralSiteCoordinate(i_sublat_before, [0, 0, 0])
            site_after = site_rep * site_before
            i_sublat_after = site_after.sublattice()
            for i_occ_before in range(len(occ_sublat_rep)):
                i_occ_after = occ_sublat_rep[i_occ_before]

                orientation_name_before = occ_dof[i_sublat_before][i_occ_before]
                orientation_name_after = occ_dof[i_sublat_after][i_occ_after]

                # assert occupants map (chemical name match)
                assert (
                    occupants[orientation_name_before].name()
                    == occupants[orientation_name_after].name()
                )

    # print(xtal.pretty_json(xtal_prim.to_dict()))

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {"occ": {"site_basis_functions": "occupation"}}
            },
        },
        verbose=False,
    )
    functions, clusters = (builder.functions, builder.clusters)

    for b in [0]:
        # "A"
        expected = np.array(
            [
                [1.0],
            ]
        )

        assert np.allclose(
            get_occ_site_functions(builder.occ_site_functions, b),
            expected,
        )

    for b in [1]:
        # "A", "mol.x", "mol.y", "mol.z"
        expected = np.array(
            [
                [1.0, 1.0, 1.0, 1.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ]
        )
        assert np.allclose(
            get_occ_site_functions(builder.occ_site_functions, b),
            expected,
        )

    for b in [2]:
        # "A", "mol.x", "mol.y", "mol.z"
        expected = np.array(
            [
                [1.0, 1.0, 1.0, 1.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ]
        )
        assert np.allclose(
            get_occ_site_functions(builder.occ_site_functions, b),
            expected,
        )

    for b in [3]:
        # "A", "mol.x", "mol.y", "mol.z"
        expected = np.array(
            [
                [1.0, 1.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 1.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
            ]
        )
        assert np.allclose(
            get_occ_site_functions(builder.occ_site_functions, b),
            expected,
        )

    # import os
    # import pathlib
    # from utils.helpers import print_expected_cluster_functions_detailed
    #
    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_site_functions_fcc_4.json",
    # )
    with open(session_shared_datadir / "expected_occ_site_functions_fcc_4.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))


def test_occ_fcc_4b(session_shared_datadir):
    """Change order of occupants

    This affects the site basis function (phi) values on that sublattice,
    but the generated functions are the same as test_occ_fcc_3a
    (because they are functions of the phi which remain in the same order).
    """

    mol_x = xtal.Occupant(
        name="mol",
        atoms=[
            xtal.AtomComponent(name="B", coordinate=[-0.1, 0.0, 0.0], properties={}),
            xtal.AtomComponent(name="B", coordinate=[0.1, 0.0, 0.0], properties={}),
        ],
    )
    mol_y = xtal.Occupant(
        name="mol",
        atoms=[
            xtal.AtomComponent(name="B", coordinate=[0.0, -0.1, 0.0], properties={}),
            xtal.AtomComponent(name="B", coordinate=[0.0, 0.1, 0.0], properties={}),
        ],
    )
    mol_z = xtal.Occupant(
        name="mol",
        atoms=[
            xtal.AtomComponent(name="B", coordinate=[0.0, 0.0, -0.1], properties={}),
            xtal.AtomComponent(name="B", coordinate=[0.0, 0.0, 0.1], properties={}),
        ],
    )
    atom_A = xtal.Occupant(
        name="A",
        atoms=[
            xtal.AtomComponent(name="A", coordinate=[0.0, 0.0, 0.0], properties={}),
        ],
    )
    occupants = {"mol.x": mol_x, "mol.y": mol_y, "mol.z": mol_z, "A": atom_A}
    occ_dof = [
        ["A"],
        ["A", "mol.x", "mol.y", "mol.z"],
        ["mol.x", "mol.y", "mol.z", "A"],
        ["A", "mol.y", "mol.z", "mol.x"],
    ]
    xtal_prim = xtal.Prim(
        lattice=xtal.Lattice(
            column_vector_matrix=np.eye(3),
        ),
        coordinate_frac=np.array(
            [
                [0.0, 0.0, 0.0],
                [0.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 0.5, 0.0],
            ]
        ).T,
        occ_dof=occ_dof,
        occupants=occupants,
    )
    prim = casmconfig.Prim(xtal_prim)
    assert len(prim.factor_group.elements) == 48
    occ_symgroup_rep = prim.occ_symgroup_rep

    for i_factor_group, occ_op_rep in enumerate(occ_symgroup_rep):
        site_rep = prim.integral_site_coordinate_symgroup_rep[i_factor_group]
        for i_sublat_before, occ_sublat_rep in enumerate(occ_op_rep):
            site_before = xtal.IntegralSiteCoordinate(i_sublat_before, [0, 0, 0])
            site_after = site_rep * site_before
            i_sublat_after = site_after.sublattice()
            for i_occ_before in range(len(occ_sublat_rep)):
                i_occ_after = occ_sublat_rep[i_occ_before]

                orientation_name_before = occ_dof[i_sublat_before][i_occ_before]
                orientation_name_after = occ_dof[i_sublat_after][i_occ_after]

                # assert occupants map (chemical name match)
                assert (
                    occupants[orientation_name_before].name()
                    == occupants[orientation_name_after].name()
                )

    # print(xtal.pretty_json(xtal_prim.to_dict()))

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {
                    "occ": {
                        "site_basis_functions": {
                            "type": "occupation",
                            "reference_occ": ["A", "A", "A", "A"],
                        }
                    }
                }
            },
        },
        verbose=False,
    )
    functions, clusters = (builder.functions, builder.clusters)

    for b in [0]:
        # "A"
        expected = np.array(
            [
                [1.0],
            ]
        )

        assert np.allclose(
            get_occ_site_functions(builder.occ_site_functions, b),
            expected,
        )

    for b in [1]:
        # "A", "mol.x", "mol.y", "mol.z"
        expected = np.array(
            [
                [1.0, 1.0, 1.0, 1.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ]
        )
        assert np.allclose(
            get_occ_site_functions(builder.occ_site_functions, b),
            expected,
        )

    for b in [2]:
        # "mol.x", "mol.y", "mol.z", "A"
        expected = np.array(
            [
                [1.0, 1.0, 1.0, 1.0],
                [0.0, 1.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
            ]
        )
        assert np.allclose(
            get_occ_site_functions(builder.occ_site_functions, b),
            expected,
        )

    for b in [3]:
        # "A", "mol.y", "mol.z", "mol.x"
        expected = np.array(
            [
                [1.0, 1.0, 1.0, 1.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ]
        )
        assert np.allclose(
            get_occ_site_functions(builder.occ_site_functions, b),
            expected,
        )

    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_site_functions_fcc_4.json",
    # )
    with open(session_shared_datadir / "expected_occ_site_functions_fcc_4.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))


def test_occ_fcc_5a(session_shared_datadir):
    """Prim with occupation DoF with Cmagspin A.up, A.down"""

    atom_A_up = xtal.Occupant(
        name="A",
        atoms=[
            xtal.AtomComponent(
                name="A",
                coordinate=[0.0, 0.0, 0.0],
                properties={"Cmagspin": [1.0]},
            ),
        ],
    )
    atom_A_down = xtal.Occupant(
        name="A",
        atoms=[
            xtal.AtomComponent(
                name="A",
                coordinate=[0.0, 0.0, 0.0],
                properties={"Cmagspin": [-1.0]},
            ),
        ],
    )
    occupants = {"A.up": atom_A_up, "A.down": atom_A_down}
    occ_dof = [
        ["A.up", "A.down"],
    ]

    xtal_prim = xtal.Prim(
        lattice=xtal.Lattice(
            column_vector_matrix=np.array(
                [
                    [0.0, 2.0, 2.0],
                    [2.0, 0.0, 2.0],
                    [2.0, 2.0, 0.0],
                ]
            ),
        ),
        coordinate_frac=np.array(
            [
                [0.0, 0.0, 0.0],
            ]
        ).T,
        occ_dof=occ_dof,
        occupants=occupants,
    )
    prim = casmconfig.Prim(xtal_prim)
    assert len(prim.factor_group.elements) == 96
    occ_symgroup_rep = prim.occ_symgroup_rep

    for i_factor_group, occ_op_rep in enumerate(occ_symgroup_rep):
        site_rep = prim.integral_site_coordinate_symgroup_rep[i_factor_group]
        for i_sublat_before, occ_sublat_rep in enumerate(occ_op_rep):
            site_before = xtal.IntegralSiteCoordinate(i_sublat_before, [0, 0, 0])
            site_after = site_rep * site_before
            i_sublat_after = site_after.sublattice()
            for i_occ_before in range(len(occ_sublat_rep)):
                i_occ_after = occ_sublat_rep[i_occ_before]

                orientation_name_before = occ_dof[i_sublat_before][i_occ_before]
                orientation_name_after = occ_dof[i_sublat_after][i_occ_after]

                # assert occupants map (chemical name match)
                assert (
                    occupants[orientation_name_before].name()
                    == occupants[orientation_name_after].name()
                )

    # print(xtal.pretty_json(xtal_prim.to_dict()))

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 4.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {"occ": {"site_basis_functions": "chebychev"}}
            },
        },
        verbose=False,
    )
    functions, clusters = (builder.functions, builder.clusters)

    for b in [0]:
        # "A.up, A.down"
        expected = np.array(
            [
                [1.0, 1.0],
                [-1.0, 1.0],
            ]
        )

        assert np.allclose(
            get_occ_site_functions(builder.occ_site_functions, b),
            expected,
        )

    from casm.bset.json_io import dump

    dump(builder.to_dict(), "cluster_functions.json", force=True)

    # import os
    # import pathlib
    # from utils.helpers import print_expected_cluster_functions_detailed
    #
    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_site_functions_fcc_5.json",
    # )
    with open(session_shared_datadir / "expected_occ_site_functions_fcc_5.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))



from typing import Optional

import numpy as np

from casm.bset.misc import (
    almost_equal,
    factor_by_mode,
    irrational_to_tex_string,
)
from casm.bset.polynomial_functions import (
    PolynomialFunction,
    Variable,
)
from libcasm.clexulator import (
    PrimNeighborList,
)


class CppFormatProperties:
    """Holds options for C++ formatting"""

    def __init__(
        self,
        coeff_fmt_spec: str = ".10f",
        coeff_atol: float = 1e-10,
    ):
        """

        .. rubric:: Constructor

        Parameters
        ----------
        coeff_fmt_spec: str
            Format specification for coefficient printing.

        coeff_atol: float
            Tolerance for checking for zeros and ones.
        """
        self.coeff_fmt_spec = coeff_fmt_spec
        """str: Format specification for coefficient printing."""

        self.coeff_atol = coeff_atol
        """float: Tolerance for checking for zeros and ones."""


def is_one(x, atol: float = 1e-10):
    return almost_equal(x, 1.0, abs_tol=atol)


def occ_func_cpp_str(
    var: Variable,
    prim_neighbor_list: PrimNeighborList,
    occupant_index_argname: str,
    mode: str = "cpp",
) -> str:
    """Return a C++ expression to evaluate an occupation site basis variable

    Parameters
    ----------
    var: casm.bset.Variable
        A :class:`casm.bset.Variable` representing the occupation site basis
        function being evaluated.
    prim_neighbor_list: libcasm.clexulator.PrimNeighborList
        The primitive neighbor list, used to find the sublattice index
    occupant_index_argname: str
        The occupant index argument name. Usually "occ_i" or "occ_f".
    mode: str = "cpp"
        The mode for the output. Options are "cpp" and "latex".

    Returns
    -------
    cpp_str: str
        C++ expression to evaluate the specified variable, with form:

        .. code-block:: Python

            # b: int, sublattice index
            # m: int, site basis function index
            cpp_str = f"m_occ_func{b}_{m}[{occupant_index_argname}]"

    """
    if var.key != "occ":
        raise Exception('Error in occ_func_cpp_str: key != "occ"')
    # occupant_index_str = "occ_i", "occ_f"
    nlist_sublat_indices = prim_neighbor_list.sublattice_indices()
    i = var.neighborhood_site_index % len(nlist_sublat_indices)
    b = nlist_sublat_indices[i]
    m = var.site_basis_function_index
    n = var.neighborhood_site_index
    if mode == "cpp":
        return f"m_occ_func_{b}_{m}[{occupant_index_argname}]"
    elif mode == "latex":
        return f"{var.name}({occupant_index_argname}_{{{n}}})"
        # return f"\\phi_{{{b},{m}}}({occupant_index_argname}_{{{n}}})"
    else:
        raise ValueError(f"Invalid mode: {mode}")


def variable_cpp_str(
    var: Variable,
    prim_neighbor_list: PrimNeighborList,
    mode: str = "cpp",
    label_site_using: str = "neighborhood_site_index",
) -> str:
    """Return a C++ expression to access an evaluated variable

    Parameters
    ----------
    var: casm.bset.polynomial_functions.Variable
        The variable to be accessed.
    prim_neighbor_list: PrimNeighborList
        The PrimNeighborList used when constructing the functions.
    mode: str = "cpp"
        The mode for the output. Options are "cpp" and "latex".
    label_site_using = "neighborhood_site_index"
        The label to use for the site index in the output. Options are
        "neighborhood_site_index" and "cluster_site_index".

    Returns
    -------
    cpp_str: str
        A C++ expression for a function that accesses the evaluated value of the
        variable.

    """
    nlist_sublat_indices = prim_neighbor_list.sublattice_indices()
    if var.key == "occ":
        # occ DoF / site basis function
        i = var.neighborhood_site_index % len(nlist_sublat_indices)
        b = nlist_sublat_indices[i]
        m = var.site_basis_function_index
        n = var.neighborhood_site_index
        cl = var.cluster_site_index
        if mode == "cpp":
            return f"occ_func_{b}_{m}({n})"
        elif mode == "latex":
            if label_site_using == "neighborhood_site_index":
                return f"{var.name}(\\vec{{r}}_{{{n}}})"
            elif label_site_using == "cluster_site_index":
                return f"{var.name}(\\vec{{r}}_{{{cl}}})"
            else:
                raise ValueError(f"Invalid label_site_using: {label_site_using}")
        else:
            raise ValueError(f"Invalid mode: {mode}")
    elif var.neighborhood_site_index is not None:
        # local continuous DoF
        c = var.component_index
        n = var.neighborhood_site_index
        cl = var.cluster_site_index
        if mode == "cpp":
            return f"{var.key}_var_{c}<Scalar>({n})"
        elif mode == "latex":
            if label_site_using == "neighborhood_site_index":
                return f"{var.name}(\\vec{{r}}_{{{n}}})"
            elif label_site_using == "cluster_site_index":
                return f"{var.name}(\\vec{{r}}_{{{cl}}})"
            else:
                raise ValueError(f"Invalid label_site_using: {label_site_using}")
        else:
            raise ValueError(f"Invalid mode: {mode}")
    else:
        # global continuous DoF
        c = var.component_index
        if mode == "cpp":
            return f"{var.key}_var<Scalar>({c})"
        elif mode == "latex":
            return f"{var.name}"
        else:
            raise ValueError(f"Invalid mode: {mode}")


def monomial_cpp_str(
    variables: list[Variable],
    coeff: float,
    monomial_exponents: np.ndarray,
    prim_neighbor_list: PrimNeighborList,
    cpp_fmt: CppFormatProperties,
    mode: str = "cpp",
    label_site_using: str = "neighborhood_site_index",
) -> str:
    """Return a C++ expression to evaluate a monomial

    Parameters
    ----------
    variables: list[casm.bset.polynomial_functions.Variable]
        The variables associated with the PolynomialFunction including this monomial.
    coeff: float
        The monomial coefficient.
    prim_neighbor_list: PrimNeighborList
        The PrimNeighborList used when constructing the functions.
    cpp_fmt: CppFormatProperties
        C++ string formatting properties.
    mode: str = "cpp"
        The mode for the output. Options are "cpp" and "latex".
    label_site_using = "neighborhood_site_index"
        The label to use for the site index in the output. Options are
        "neighborhood_site_index" and "cluster_site_index".

    Returns
    -------
    cpp_str: str
        A C++ expression to evaluate the monomial and multiply by the coefficient.

    """
    if (monomial_exponents == np.zeros(monomial_exponents.shape)).all():
        return "1"

    use_mul = False
    cpp_str = ""
    if not is_one(coeff, atol=cpp_fmt.coeff_atol):
        cpp_str += f"{coeff:{cpp_fmt.coeff_fmt_spec}}"
        use_mul = True
    for i_var, x in enumerate(monomial_exponents):
        if x == 0:
            continue
        v = variable_cpp_str(
            variables[i_var],
            prim_neighbor_list,
            mode=mode,
            label_site_using=label_site_using,
        )
        if use_mul:
            if mode == "cpp":
                op = "*"
            elif mode == "latex":
                op = " "
            else:
                raise ValueError(f"Invalid mode: {mode}")
        else:
            op = ""
            use_mul = True
        if x == 1:
            cpp_str += f"{op}{v}"
        else:
            if mode == "cpp":
                cpp_str += f"{op}pow({v},{x})"
            elif mode == "latex":
                cpp_str += f"{op}{v}^{{{x}}}"
            else:
                raise ValueError(f"Invalid mode: {mode}")
    return cpp_str


def polynomial_sum_cpp_str(
    functions: list[PolynomialFunction],
    normalization: float,
    prim_neighbor_list: PrimNeighborList,
    cpp_fmt: CppFormatProperties,
    mode: str = "cpp",
    label_site_using: str = "neighborhood_site_index",
) -> str:
    """Return a C++ expression to evaluate a polynomial expression

    Parameters
    ----------
    functions: list[PolynomialFunction]
        The symmetrically equivalent cluster functions (i.e. that have the same
        expansion coefficient) associated with a single unit cell that need to be
        evaluated.
    normalization: float
        The normalization constant.
    prim_neighbor_list: PrimNeighborList
        The PrimNeighborList used when constructing the functions.
    cpp_fmt: CppFormatProperties
        C++ string formatting properties.
    mode: str = "cpp"
        The mode for the output. Options are "cpp" and "latex".
    label_site_using = "neighborhood_site_index"
        The label to use for the site index in the output. Options are
        "neighborhood_site_index" and "cluster_site_index".

    Returns
    -------
    cpp_str: str
        A C++ expression to evaluate the sum of polynomial functions.

    """
    if len(functions) == 0:
        return "0."

    prefix_coeff = []
    factored_coeff = []
    for function in functions:
        _prefix_coeff, _factored_coeff = factor_by_mode(function.coeff.data)
        prefix_coeff.append(_prefix_coeff)
        factored_coeff.append(_factored_coeff)

    common_prefix, factored_prefixes = factor_by_mode(np.array(prefix_coeff))

    cpp_str = ""
    if not is_one(common_prefix, atol=cpp_fmt.coeff_atol):
        if mode == "cpp":
            cpp_str += f"{common_prefix:{cpp_fmt.coeff_fmt_spec}} * "
        elif mode == "latex":
            limit = len(functions[0].variables) ** 2
            max_pow = 2
            common_prefix_tex = irrational_to_tex_string(
                common_prefix, limit=limit, max_pow=max_pow, abs_tol=1e-5
            )
            cpp_str += f"{common_prefix_tex} "
    if len(functions) > 1:
        cpp_str += "(\n"

    indent = "  "
    orbit_use_plus = False
    for i_func, function in enumerate(functions):
        if len(functions) > 1:
            cpp_str += indent * 2
        if orbit_use_plus:
            cpp_str += " + "
        else:
            orbit_use_plus = True
        if not is_one(factored_prefixes[i_func], atol=cpp_fmt.coeff_atol):
            cpp_str += f"{factored_prefixes[i_func]:{cpp_fmt.coeff_fmt_spec}} * "
        if len(function.monomial_exponents) > 1:
            cpp_str += "("
        cluster_use_plus = False
        for i_monomial, _monomial_exponents in enumerate(function.monomial_exponents):
            if cluster_use_plus:
                op = " + "
            else:
                op = ""
                cluster_use_plus = True
            cpp_str += op + monomial_cpp_str(
                variables=function.variables,
                coeff=factored_coeff[i_func][i_monomial],
                monomial_exponents=_monomial_exponents,
                prim_neighbor_list=prim_neighbor_list,
                cpp_fmt=cpp_fmt,
                mode=mode,
                label_site_using=label_site_using,
            )
        if len(function.monomial_exponents) > 1:
            cpp_str += ")"
        if len(functions) > 1:
            cpp_str += "\n"

    if len(functions) > 1:
        cpp_str += indent * 1 + ")"
    if not is_one(normalization, atol=cpp_fmt.coeff_atol):
        cpp_str += f" / {normalization}"
    return cpp_str


def orbit_bfunc_cpp_str(
    orbit_functions: list[PolynomialFunction],
    orbit_size: int,
    prim_neighbor_list: PrimNeighborList,
    cpp_fmt: CppFormatProperties,
    mode: str = "cpp",
    label_site_using: str = "neighborhood_site_index",
) -> str:
    """Return a C++ expression to evaluate a single orbit basis function

    Parameters
    ----------
    orbit_functions: list[PolynomialFunction]
        The symmetrically equivalent cluster functions (i.e. that have the same
        expansion coefficient) associated with a single unit cell that need to be
        evaluated.
    orbit_size: int
        The number of clusters in the orbit, per unit cell. Used for normalization.
    prim_neighbor_list: PrimNeighborList
        The PrimNeighborList used when constructing the functions.
    cpp_fmt: CppFormatProperties
        C++ string formatting properties.
    mode: str = "cpp"
        The mode for the output. Options are "cpp" and "latex".
    label_site_using = "neighborhood_site_index"
        The label to use for the site index in the output. Options are
        "neighborhood_site_index" and "cluster_site_index".

    Returns
    -------
    cpp_str: str
        A C++ expression to evaluate the sum of polynomial functions involved in the
        calculation of the contribution to the correlations from one unit cell (for a
        periodic cluster expansion) or for one local-cluster correlation value.

    """
    return polynomial_sum_cpp_str(
        functions=orbit_functions,
        normalization=float(orbit_size),
        prim_neighbor_list=prim_neighbor_list,
        cpp_fmt=cpp_fmt,
        mode=mode,
        label_site_using=label_site_using,
    )


def site_bfunc_cpp_str(
    point_functions: list[PolynomialFunction],
    orbit_size: int,
    prim_neighbor_list: PrimNeighborList,
    cpp_fmt: CppFormatProperties,
    mode: str = "cpp",
    label_site_using: str = "neighborhood_site_index",
) -> Optional[str]:
    """Return a C++ expression to evaluate a single point correlation function

    Parameters
    ----------
    point_functions: list[PolynomialFunction]
        The symmetrically equivalent cluster functions (i.e. that have the same
        expansion coefficient) that include the site at which the point correlations are
        evaluated.
    orbit_size: int
        The number of clusters in the orbit, per unit cell. Used for normalization.
    prim_neighbor_list: PrimNeighborList
        The PrimNeighborList used when constructing the functions.
    cpp_fmt: CppFormatProperties
        C++ string formatting properties.
    mode: str = "cpp"
        The mode for the output. Options are "cpp" and "latex".

    Returns
    -------
    cpp_str: str
        A C++ expression to evaluate a single point correlation function (for either
        periodic or local-cluster expansions).

    """
    if len(point_functions) == 0:
        return None
    return polynomial_sum_cpp_str(
        functions=point_functions,
        normalization=float(orbit_size),
        prim_neighbor_list=prim_neighbor_list,
        cpp_fmt=cpp_fmt,
        mode=mode,
        label_site_using=label_site_using,
    )


def occ_delta_site_bfunc_cpp_str(
    neighbor_list_index: int,
    point_functions: list[PolynomialFunction],
    orbit_size: int,
    prim_neighbor_list: PrimNeighborList,
    cpp_fmt: CppFormatProperties,
    mode: str = "cpp",
    label_site_using: str = "neighborhood_site_index",
) -> Optional[str]:
    """Return a C++ expression to evaluate the change in a single point correlation \
    function due to a change in occupation

    Parameters
    ----------
    neighbor_list_index: int
        The index in the prim neighbor list of the site with changing occupation.
    point_functions: list[PolynomialFunction]
        The symmetrically equivalent cluster functions (i.e. that have the same
        expansion coefficient) that include the site at which the point correlations are
        evaluated.
    orbit_size: int
        The number of clusters in the orbit, per unit cell. Used for normalization.
    prim_neighbor_list: PrimNeighborList
        The PrimNeighborList used when constructing the functions.
    cpp_fmt: CppFormatProperties
        C++ string formatting properties.
    mode: str = "cpp"
        The mode for the output. Options are "cpp" and "latex".

    Returns
    -------
    cpp_str: str
        A C++ expression to evaluate the change in a single point correlation function
        due to a change in occupation (for either periodic or local-cluster expansions).
"""
    if len(point_functions) == 0:
        return None

    def find_occ_site_var(monomial: PolynomialFunction):
        variables = monomial.variables
        if len(monomial.monomial_exponents) != 1:
            raise Exception(
                "Error in occ_delta_site_bfunc_cpp_str: "
                "failed to split point function into monomials"
            )

        occ_site_var = None
        for i_var, x in enumerate(monomial.monomial_exponents[0]):
            if x == 0:
                continue
            var = variables[i_var]
            if var.key != "occ":
                continue
            if var.neighborhood_site_index == neighbor_list_index:
                if occ_site_var is not None:
                    raise Exception(
                        "Error in occ_delta_site_bfunc_cpp_str: "
                        ">1 occ site basis function found in monomial"
                    )
                occ_site_var = var
        return occ_site_var

    monomials = []
    for point_function in point_functions:
        monomials += point_function.monomials()

    occ_delta_functions = {}
    for monomial in monomials:
        var = find_occ_site_var(monomial)
        if var is not None:
            if var not in occ_delta_functions:
                occ_delta_functions[var] = []
            occ_delta_functions[var].append(monomial / var)

    if len(occ_delta_functions) == 0:
        return None

    indent = "  "
    cpp_str = ""
    use_plus = False
    if mode == "cpp":
        occupant_index_argname_final = "occ_f"
        occupant_index_argname_init = "occ_i"
    elif mode == "latex":
        occupant_index_argname_final = "\\vec{r}^{\\ f}"
        occupant_index_argname_init = "\\vec{r}^{\\ i}"
    else:
        raise ValueError(f"Invalid mode: {mode}")

    for var, factored_point_functions in occ_delta_functions.items():
        site_var_final = occ_func_cpp_str(
            var=var,
            prim_neighbor_list=prim_neighbor_list,
            occupant_index_argname=occupant_index_argname_final,
            mode=mode,
        )
        site_var_init = occ_func_cpp_str(
            var=var,
            prim_neighbor_list=prim_neighbor_list,
            occupant_index_argname=occupant_index_argname_init,
            mode=mode,
        )
        if use_plus:
            cpp_str += "\n\n" + indent * 1 + " + "
        else:
            use_plus = True
        cpp_str += f"({site_var_final} - {site_var_init}) * "
        cpp_str += polynomial_sum_cpp_str(
            functions=factored_point_functions,
            normalization=float(orbit_size),
            prim_neighbor_list=prim_neighbor_list,
            cpp_fmt=cpp_fmt,
            mode=mode,
            label_site_using=label_site_using,
        )
    return cpp_str



import gzip
import json
import os
import pathlib
import tarfile
import typing

import libcasm.xtal as xtal


def pretty_json(
    data: dict,
) -> str:
    """Pretty-print Python dict as a JSON string."""
    return xtal.pretty_json(data)


def printpathstr(path):
    abspath = pathlib.Path(path).resolve()
    try:
        return str(abspath.relative_to(pathlib.Path.cwd()))
    except ValueError:
        return str(abspath)


def read_required(path: pathlib.Path, gz: bool = False):
    path = pathlib.Path(path)
    if path.exists():
        if gz is True:
            with gzip.open(path, "r") as f:
                return json.loads(f.read().decode("utf-8"))
        else:
            with open(path, "r") as f:
                return json.load(f)
    else:
        raise Exception("Required file: '" + printpathstr(path) + "' does not exist")


def read_contents(
    parent_dir: pathlib.Path,
    relpath: pathlib.Path,
    default: typing.Any = None,
    quiet: bool = False,
):
    """Read file contents, whether it is an individual or inside an archive

    Parameters
    ----------
    parent_dir: pathlib.Path
        The file path to a Monte Carlo run path directory

    relpath: pathlib.Path
        Relative path, referenced to parent_dir, of a file to open. The `parent_dir`
        directory may be tar gzipped, in which case this file is read from the archive.

        If `relpath.suffix == ".gz"`, the file is read as a gzipped file.

    Returns
    -------
    data: Any
        The JSON contents of the file, or a default value.
    """
    parent_dir = pathlib.Path(parent_dir)
    relpath = pathlib.Path(relpath)
    datafile_path = parent_dir / relpath
    tgz_path = parent_dir.parent / (parent_dir.name + ".tgz")

    try:
        if datafile_path.exists():
            if relpath.suffix == ".gz":
                # gzipped JSON file
                with gzip.open(datafile_path, "rb") as f:
                    return json.loads(f.read().decode("utf-8"))
            else:
                # JSON file
                with open(datafile_path, "r") as f:
                    return json.load(f)
        elif tgz_path.exists():
            with tarfile.open(tgz_path, "r:gz") as f:
                membername = str(pathlib.Path(parent_dir.name) / relpath)
                with f.extractfile(membername) as g:
                    if relpath.suffix == ".gz":
                        # gzipped JSON file
                        return json.loads(gzip.open(g).read().decode("utf-8"))
                    else:
                        # JSON file
                        return json.load(g)
        else:
            if not quiet:
                print(printpathstr(datafile_path) + ": does not exist, skipping")
            return default
    except Exception:
        if not quiet:
            print(printpathstr(datafile_path) + ": error reading, skipping")
        return default


def read_optional(path: pathlib.Path, default: typing.Any = None, gz: bool = False):
    path = pathlib.Path(path)
    if path.exists():
        if gz is True:
            with gzip.open(path, "r") as f:
                return json.loads(f.read().decode("utf-8"))
        else:
            with open(path, "r") as f:
                return json.load(f)
    return default


def read_cascading(paths: list[pathlib.Path], quiet: bool = False, gz: bool = False):
    """Find a required file that may be at multiple locations"""
    for path in paths:
        data = read_optional(path, gz=gz)
        if data is not None:
            if not quiet:
                print("reading:", os.path.relpath(path))
            return data
    print("Required file not found. Checked:")
    for path in paths:
        print("- " + printpathstr(path) + ": does not exist")
    raise Exception("Required file does not exist")


def dump(
    data, path: pathlib.Path, force: bool = False, quiet: bool = False, gz: bool = False
):
    """Json dump with overwrite/skipping/write output messaging"""

    def _write(data, path: pathlib.Path, gz: bool = False):
        path.parent.mkdir(parents=True, exist_ok=True)
        if gz is True:
            with gzip.open(path, "w") as f:
                f.write(xtal.pretty_json(data).encode("utf-8"))
        else:
            with open(path, "w") as f:
                f.write(xtal.pretty_json(data))

    path = pathlib.Path(path)
    if path.exists():
        if force:
            if not quiet:
                print("overwrite:", printpathstr(path))
            _write(data, path, gz=gz)
        elif not quiet:
            print("skipping:", printpathstr(path))
    else:
        if not quiet:
            print("write:", printpathstr(path))
        _write(data, path, gz=gz)


def safe_dump(
    data,
    path: pathlib.Path,
    force: bool = False,
    quiet: bool = False,
    gz: bool = False,
):
    """Json dump with overwrite/skipping/write output messaging

    Writes to the temporary file `path + ".tmp"`, then removes `path`
    and renames the temporary file, to avoid losing the original file
    without writing the new file. This method does not avoid race
    conditions.

    If the temporary file already exists an exception is raised.
    """
    path = pathlib.Path(path)

    def _safe_write(data, path, gz: bool = False):
        path.parent.mkdir(parents=True, exist_ok=True)
        tmp_path = pathlib.Path(str(path) + ".tmp")
        if tmp_path.exists():
            raise Exception("Error: " + str(tmp_path) + " already exists")

        if gz is True:
            with gzip.open(tmp_path, "w") as f:
                f.write(xtal.pretty_json(data).encode("utf-8"))
        else:
            with open(tmp_path, "w") as f:
                f.write(xtal.pretty_json(data))

        if path.exists():
            path.unlink()
        tmp_path.rename(path)

    if path.exists():
        if force:
            if not quiet:
                print("overwrite:", printpathstr(path))
            _safe_write(data, path, gz=gz)
        elif not quiet:
            print("skipping:", printpathstr(path))
    else:
        if not quiet:
            print("write:", printpathstr(path))
        _safe_write(data, path, gz=gz)


def get(data: typing.Union[dict, list], path: list, default: typing.Any = None):
    """Get item from a heirarchical data structure of dict and lists

    Parameters
    ----------
    data: Union[dict, list]
        A heirarchical data structure of dict and lists
    path: list
        A path of keys and indices into the data
    default: Any = None
        A default value to return if a key does not exist. Invalid indices into
        lists raise Exceptions instead of returning the default value.
    """
    if isinstance(data, dict):
        key = str(path[0])
        if key in data:
            if len(path) == 1:
                return data[key]
            else:
                return get(data[key], path=path[1:], default=default)
        else:
            return default
    elif isinstance(data, list):
        index = int(path[0])
        if index < 0:
            raise Exception("Error in json_io.get: index < 0")
        elif index >= len(data):
            raise Exception("Error in json_io.get: index > len(data)")
        else:
            if len(path) == 1:
                return data[index]
            else:
                return get(data[index], path=path[1:], default=default)
    else:
        raise Exception("Error in json_io.get: not a dict or list")



import json

# from utils.expected_disp_functions import (
#     expected_occ_functions_fcc_1,
#     expected_occ_functions_hcp_1,
#     expected_occ_functions_lowsym_1,
# )
from utils.helpers import (
    assert_expected_cluster_functions_detailed,
)

import libcasm.clusterography as casmclust
import libcasm.configuration as casmconfig
import libcasm.occ_events as occ_events
import libcasm.xtal.prims as xtal_prims
from casm.bset import (
    build_cluster_functions,
)
from casm.bset.cluster_functions import (
    BasisFunctionSpecs,
    ClexBasisSpecs,
    make_neighborhood,
    make_occevent_cluster_specs,
)


def test_occ_fcc_1a(session_shared_datadir):
    """build_functions with libcasm.xtal.Prim / dict bspecs"""
    xtal_prim = xtal_prims.FCC(
        r=0.5,
        occ_dof=["A", "B", "C"],
    )
    # print(xtal.pretty_json(xtal_prim.to_dict()))

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {"occ": {"site_basis_functions": "occupation"}}
            },
        },
    )
    functions, clusters = (builder.functions, builder.clusters)

    # import os
    # import pathlib
    # from utils.helpers import print_expected_cluster_functions_detailed
    #
    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_functions_fcc_1.json",
    # )
    with open(session_shared_datadir / "expected_occ_functions_fcc_1.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))


def test_occ_fcc_1b(session_shared_datadir):
    """build_functions with libcasm.configuration.Prim / ClexBasisSpecs"""
    xtal_prim = xtal_prims.FCC(
        r=0.5,
        occ_dof=["A", "B", "C"],
    )

    prim = casmconfig.Prim(xtal_prim)
    builder = build_cluster_functions(
        prim=prim,
        clex_basis_specs=ClexBasisSpecs(
            cluster_specs=casmclust.ClusterSpecs(
                xtal_prim=prim.xtal_prim,
                generating_group=prim.factor_group,
                max_length=[0.0, 0.0, 1.01, 1.01],
            ),
            basis_function_specs=BasisFunctionSpecs(
                dof_specs={"occ": {"site_basis_functions": "occupation"}},
            ),
        ),
    )
    functions, clusters = (builder.functions, builder.clusters)

    # import os
    # import pathlib
    # from utils.helpers import print_expected_cluster_functions_detailed
    #
    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_functions_fcc_1.json",
    # )
    with open(session_shared_datadir / "expected_occ_functions_fcc_1.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))


def test_occ_fcc_local_1(session_shared_datadir):
    xtal_prim = xtal_prims.FCC(
        a=1.0,
        occ_dof=["A", "B", "Va"],
    )
    prim = casmconfig.Prim(xtal_prim)
    # print(xtal.pretty_json(xtal_prim.to_dict()))

    occ_system = occ_events.OccSystem(xtal_prim=prim.xtal_prim)

    occevent = occ_events.OccEvent.from_dict(
        data={
            "trajectories": [
                [
                    {"coordinate": [0, 0, 0, 0], "occupant_index": 0},
                    {"coordinate": [0, 1, 0, 0], "occupant_index": 0},
                ],
                [
                    {"coordinate": [0, 1, 0, 0], "occupant_index": 2},
                    {"coordinate": [0, 0, 0, 0], "occupant_index": 2},
                ],
            ]
        },
        system=occ_system,
    )

    cluster_specs = make_occevent_cluster_specs(
        prim=prim,
        phenomenal_occ_event=occevent,
        max_length=[0.0, 0.0, 1.01],
        cutoff_radius=[0.0, 1.01, 1.01],
    )
    assert cluster_specs.generating_group().head_group is prim.factor_group
    orbits = cluster_specs.make_orbits()

    expected_neighborhood_size = 26
    assert len(make_neighborhood(clusters=orbits)) == expected_neighborhood_size

    builder = build_cluster_functions(
        prim=prim,
        clex_basis_specs=ClexBasisSpecs(
            cluster_specs=cluster_specs,
            basis_function_specs=BasisFunctionSpecs(
                dof_specs={"occ": {"site_basis_functions": "occupation"}},
            ),
        ),
    )
    functions, clusters = (builder.functions, builder.clusters)
    equivalent_functions, equivalent_clusters = (
        builder.equivalent_functions,
        builder.equivalent_clusters,
    )

    expected_n_clex = 6
    assert len(equivalent_functions) == expected_n_clex
    assert len(equivalent_clusters) == expected_n_clex

    assert len(make_neighborhood(clusters=clusters)) == expected_neighborhood_size
    for _clusters in equivalent_clusters:
        assert len(make_neighborhood(clusters=_clusters)) == expected_neighborhood_size

    # import os
    # import pathlib
    # from utils.helpers import print_expected_cluster_functions_detailed
    #
    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_functions_fcc_local_1_prototype.json",
    # )
    with open(
        session_shared_datadir / "expected_occ_functions_fcc_local_1_prototype.json"
    ) as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))
    for i_clex, _functions in enumerate(equivalent_functions):
        _clusters = equivalent_clusters[i_clex]
        # print_expected_cluster_functions_detailed(
        #     _functions,
        #     file=pathlib.Path(os.path.realpath(__file__)).parent
        #     / "data"
        #     / f"expected_occ_functions_fcc_local_1_equiv_{i_clex}.json",
        # )
        with open(
            session_shared_datadir
            / f"expected_occ_functions_fcc_local_1_equiv_{i_clex}.json"
        ) as f:
            assert_expected_cluster_functions_detailed(
                _functions, _clusters, json.load(f)
            )


def test_occ_hcp_1(session_shared_datadir):
    xtal_prim = xtal_prims.HCP(
        r=0.5,
        occ_dof=["A", "B", "C"],
    )

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {"occ": {"site_basis_functions": "chebychev"}}
            },
        },
    )
    functions, clusters = (builder.functions, builder.clusters)

    # import os
    # import pathlib
    # from utils.helpers import print_expected_cluster_functions_detailed
    #
    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_functions_hcp_1.json",
    # )
    with open(session_shared_datadir / "expected_occ_functions_hcp_1.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))


def test_occ_lowsym_1(lowsym_occ_prim, session_shared_datadir):
    xtal_prim = lowsym_occ_prim

    builder = build_cluster_functions(
        prim=xtal_prim,
        clex_basis_specs={
            "cluster_specs": {
                "orbit_branch_specs": {
                    "2": {"max_length": 1.01},
                    "3": {"max_length": 1.01},
                },
            },
            "basis_function_specs": {
                "dof_specs": {"occ": {"site_basis_functions": "chebychev"}}
            },
        },
    )
    functions, clusters = (builder.functions, builder.clusters)

    # import os
    # import pathlib
    # from utils.helpers import print_expected_cluster_functions_detailed
    #
    # print_expected_cluster_functions_detailed(
    #     functions,
    #     file=pathlib.Path(os.path.realpath(__file__)).parent
    #     / "data"
    #     / "expected_occ_functions_lowsym_1.json",
    # )
    with open(session_shared_datadir / "expected_occ_functions_lowsym_1.json") as f:
        assert_expected_cluster_functions_detailed(functions, clusters, json.load(f))
